import os
import tempfile
import pytest
from ..fitslib import read_header, FITS
from ..fits_exceptions import FITSFormatError


def test_header_junk():
    """
    test lenient treatment of garbage written by IDL mwrfits
    """

    data = b"""SIMPLE  =                    T /Primary Header created by MWRFITS v1.11         BITPIX  =                   16 /                                                NAXIS   =                    0 /                                                EXTEND  =                    T /Extensions may be present                       BLAT    =                    1 /integer                                         FOO     =              1.00000 /float (or double?)                              BAR@    =                  NAN /float NaN                                       BI.Z    =                  NaN /double NaN                                      BAT     =                  INF /1.0 / 0.0                                       BOO     =                 -INF /-1.0 / 0.0                                      QUAT    = '        '           /blank string                                    QUIP    = '1.0     '           /number in quotes                                QUIZ    = ' 1.0    '           /number in quotes with a leading space           QUI\xf4\x04   = 'NaN     '           /NaN in quotes                                   HIERARCH QU.@D = 'Inf     '                                                     END                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             """  # noqa

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with open(fname, 'wb') as fobj:
            fobj.write(data)

        h = read_header(fname)
        # these keys are not hierarch but we can parse the name and then
        # leave the value as a string, so we do that.
        assert h['bar@'] == 'NAN', 'NAN garbage'
        assert h['bi.z'] == 'NaN', 'NaN garbage'
        assert h['bat'] == 'INF', 'INF garbage'
        assert h['boo'] == '-INF', '-INF garbage'
        assert h['quat'] == '', 'blank'
        assert h['quip'] == '1.0', '1.0 in quotes'
        assert h['quiz'] == ' 1.0', '1.0 in quotes'
        # the key in the header is 'QUI' + two non-ascii chars and gets
        # translated to `QUI__`
        assert h['qui__'] == 'NaN', 'NaN in quotes'
        # this key is `HIERARCH QU.@D` in the header and so gets read as is
        assert h['qu.@d'] == 'Inf', 'Inf in quotes'


def test_Header_Junk_Non_Ascii():
    data = b"SIMPLE  =                    T / file does conform to FITS standard             BITPIX  =                   16 / number of bits per data pixel                  NAXIS   =                    0 / number of data axes                            EXTEND  =                    T / FITS dataset may contain extensions            COMMENT   FITS (Flexible Image Transport System) format is defined in 'AstronomyCOMMENT   and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H @\x0f@\x0f \x02\x05\x18@\x02\x02\xc5@\x0c\x03\xf3@\x080\x02\x03\xbc@\x0f@@@@@@@@                                                END                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             "  # noqa

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')
        with open(fname, 'wb') as fobj:
            fobj.write(data)

        h = read_header(fname)
        assert h["@_@_"] is None


def test_missing_xtension_keyword():
    """
    Misformatted header with extension not properly marked with
    XTENSION
    """

    data = b"""SIMPLE  =                    T / This is a FITS file                            BITPIX  =                    8 /                                                NAXIS   =                    0 /                                                EXTEND  =                    T / This file may contain FITS extensions          NEXTEND =                    7 / Number of extensions                           END                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             SIMPLE  =                    T / file does conform to FITS standard             BITPIX  =                   32 / number of bits per data pixel                  NAXIS   =                    2 / number of data axes                            NAXIS1  =                   30 / length of data axis 1                          NAXIS2  =                   30 / length of data axis 2                          EXTEND  =                    T / FITS dataset may contain extensions            END                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             """  # noqa

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with open(fname, 'wb') as fobj:
            fobj.write(data)

        with pytest.raises(FITSFormatError):
            with FITS(fname) as fits:
                print(fits)



"""
utilities for the fits library
"""

from contextlib import contextmanager
import sys
import numpy

from . import _fitsio_wrap

if sys.version_info >= (3, 0, 0):
    IS_PY3 = True
else:
    IS_PY3 = False

_FLOATING_NULL_VALUE = _fitsio_wrap.cfitsio_null_value_for_nan()


class FITSRuntimeWarning(RuntimeWarning):
    pass


def cfitsio_version(asfloat=False):
    """
    Return the cfitsio version as a string.
    """
    # use string version to avoid roundoffs
    ver = '%0.3f' % _fitsio_wrap.cfitsio_version()
    if asfloat:
        return float(ver)
    else:
        return ver


def cfitsio_is_bundled():
    """Return True if library was built with a
    bundled copy of cfitsio.
    """
    return _fitsio_wrap.cfitsio_is_bundled()


if sys.version_info > (3, 0, 0):
    _itypes = (int,)
    _stypes = (str, bytes)
else:
    _itypes = (int, long)  # noqa - only for py2
    _stypes = (
        basestring,  # noqa - only for py2
        unicode,  # noqa - only for py2
    )  # noqa - only for py2

_itypes += (
    numpy.uint8,
    numpy.int8,
    numpy.uint16,
    numpy.int16,
    numpy.uint32,
    numpy.int32,
    numpy.uint64,
    numpy.int64,
)

# different for py3
if numpy.lib.NumpyVersion(numpy.__version__) < "1.28.0":
    _stypes += (
        numpy.string_,
        numpy.str_,
    )
else:
    _stypes += (
        numpy.bytes_,
        numpy.str_,
    )

# for header keywords
_ftypes = (float, numpy.float32, numpy.float64)


def isstring(arg):
    return isinstance(arg, _stypes)


def isinteger(arg):
    return isinstance(arg, _itypes)


def is_object(arr):
    if arr.dtype.descr[0][1][1] == 'O':
        return True
    else:
        return False


def fields_are_object(arr):
    isobj = numpy.zeros(len(arr.dtype.names), dtype=bool)
    for i, name in enumerate(arr.dtype.names):
        if is_object(arr[name]):
            isobj[i] = True
    return isobj


def is_little_endian(array):
    """
    Return True if array is little endian, False otherwise.

    Parameters
    ----------
    array: numpy array
        A numerical python array.

    Returns
    -------
    Truth value:
        True for little-endian

    Notes
    -----
    Strings are neither big or little endian.  The input must be a simple numpy
    array, not an array with fields.
    """
    if numpy.little_endian:
        machine_little = True
    else:
        machine_little = False

    byteorder = array.dtype.base.byteorder
    return (byteorder == '<') or (machine_little and byteorder == '=')


def array_to_native(array, inplace=False):
    """
    Convert an array to the native byte order.

    NOTE: the inplace keyword argument is not currently used.
    """
    if numpy.little_endian:
        machine_little = True
    else:
        machine_little = False

    data_little = False
    if array.dtype.names is None:
        if array.dtype.base.byteorder == '|':
            # strings and 1 byte integers
            return array

        data_little = is_little_endian(array)
    else:
        # assume all are same byte order: we only need to find one with
        # little endian
        for fname in array.dtype.names:
            if is_little_endian(array[fname]):
                data_little = True
                break

    if (machine_little and not data_little) or (
        not machine_little and data_little
    ):
        output = array.byteswap(inplace)
    else:
        output = array

    return numpy.require(output, requirements=['ALIGNED'])


if numpy.lib.NumpyVersion(numpy.__version__) >= "2.0.0":
    copy_if_needed = None
elif numpy.lib.NumpyVersion(numpy.__version__) < "1.28.0":
    copy_if_needed = False
else:
    # 2.0.0 dev versions, handle cases where copy may or may not exist
    try:
        numpy.array([1]).__array__(copy=None)
        copy_if_needed = None
    except TypeError:
        copy_if_needed = False


def array_to_native_c(array_in, inplace=False):
    # copy only made if not C order
    arr = numpy.require(
        array_in,
        requirements=['C_CONTIGUOUS', 'ALIGNED'],
    )
    return array_to_native(arr, inplace=inplace)


def mks(val):
    """
    make sure the value is a string, paying mind to python3 vs 2
    """
    if sys.version_info > (3, 0, 0):
        if isinstance(val, bytes):
            sval = str(val, 'utf-8')
        else:
            sval = str(val)
    else:
        sval = str(val)

    return sval


@contextmanager
def _nonfinite_as_cfitsio_floating_null_value(data, target_hdu_compressed):
    try:
        has_nonfinite = False
        if (
            data is not None
            and data.dtype.kind == "f"
            and target_hdu_compressed
        ):
            msk_nonfinite = ~numpy.isfinite(data)
            if numpy.any(msk_nonfinite):
                has_nonfinite = True
                old_vals = data[msk_nonfinite]
                data[msk_nonfinite] = _FLOATING_NULL_VALUE

        yield data, has_nonfinite
    finally:
        if has_nonfinite:
            data[msk_nonfinite] = old_vals



import os
import argparse


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--moddir', required=True,
                        help='directory containing modified files')
    parser.add_argument('--dir', required=True,
                        help='directory containing unmodified files')
    parser.add_argument('--patch-dir', required=True)
    return parser.parse_args()


def main():
    args = get_args()

    os.makedirs(args.patch_dir, exist_ok=True)

    for root, _, files in os.walk(args.dir):
        for fname in files:
            src = os.path.join(args.dir, fname)
            dst = os.path.join(args.moddir, fname)
            patch = os.path.join(args.patch_dir, fname + '.patch')
            os.system('diff -u %s %s > %s' % (src, dst, patch))
            with open(patch, 'rb') as fp:
                buff = fp.read()
            if len(buff) == 0:
                os.remove(patch)
            else:
                print(fname)
        break


main()



import os
import tempfile
import numpy as np
import fitsio


def test_compression_nocompress():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        img = np.ones((20, 20))
        with fitsio.FITS(fn, 'rw', clobber=True) as fits:
            fits.write(img)
        with fitsio.FITS(fn) as fits:
            assert len(fits) == 1


def test_compression_diskfile_kwargs():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        img = np.ones((20, 20))
        with fitsio.FITS(fn, 'rw', clobber=True) as fits:
            fits.write(
                img,
                compress='RICE',
                tile_dims=(10, 5),
                qlevel=7.0,
                qmethod='SUBTRACTIVE_DITHER_2',
                dither_seed=42,
            )
        with fitsio.FITS(fn) as fits:
            assert len(fits) == 2
        hdr = fitsio.read_header(fn, ext=1)
        for key, val in [
            ('ZTILE1', 5),
            ('ZTILE2', 10),
            ('ZQUANTIZ', 'SUBTRACTIVE_DITHER_2'),
            ('ZDITHER0', 42),
            ('ZCMPTYPE', 'RICE_ONE'),
        ]:
            assert hdr[key] == val


def test_compression_efns():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        img = np.ones((20, 20))
        with fitsio.FITS(fn + '[compress G]', 'rw', clobber=True) as fits:
            fits.write(img)
        hdr = fitsio.read_header(fn, ext=1)
        for key, val in [
            ('ZTILE1', 20),
            ('ZTILE2', 1),
            ('ZQUANTIZ', 'SUBTRACTIVE_DITHER_1'),
            ('ZCMPTYPE', 'GZIP_1'),
        ]:
            assert hdr[key] == val


def test_compression_efns_kwargs():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        img = np.ones((20, 20))
        with fitsio.FITS(
            fn + '[compress G 5 10; qz 8.0]', 'rw', clobber=True
        ) as fits:
            fits.write(img, dither_seed=42)
        hdr = fitsio.read_header(fn, ext=1)
        for key, val in [
            ('ZTILE1', 5),
            ('ZTILE2', 10),
            ('ZQUANTIZ', 'SUBTRACTIVE_DITHER_2'),
            ('ZCMPTYPE', 'GZIP_1'),
            ('ZDITHER0', 42),
        ]:
            assert hdr[key] == val


def test_compression_qlevels_none_zero():
    default_kws = {
        "compress": fitsio.GZIP_2,
        "tile_dims": np.array([100, 100]),
        "qmethod": fitsio.SUBTRACTIVE_DITHER_2,
    }
    with tempfile.TemporaryDirectory() as tmpdir:
        fnpat = os.path.join(tmpdir, 'test-%i.fits')

        H, W = 200, 200
        bigimg = np.random.uniform(size=(H, W))
        results = []
        # None: don't even use compression at all
        # 0: lossless gzip
        for i, qlevel in enumerate([None, 0, 16, 4, 1]):
            fn = fnpat % i
            ql = qlevel
            kw = {}
            kw.update(default_kws)
            if ql is None:
                kw.update(compress=0)
                ql = 0
            kw["qlevel"] = ql
            with fitsio.FITS(fn, 'rw', clobber=True) as fits:
                fits.write(bigimg, dither_seed=42, **kw)
            filesize = os.stat(fn).st_size
            img2 = fitsio.read(fn)
            rms = np.sqrt(np.mean((img2 - bigimg) ** 2))
            results.append((qlevel, filesize, rms))
        # No compression
        q, sz, rms = results[0]
        assert sz == 2880 * (1 + int(np.ceil(H * W * 8 / 2880.0)))
        assert rms == 0.0
        # GZIP lossless
        q, sz, rms = results[1]
        assert rms == 0.0
        # Decreasing file size
        for r1, r2 in zip(results, results[1:]):
            q1, sz1, rms1 = r1
            q2, sz2, rms2 = r2
            assert sz1 > sz2
            assert rms1 <= rms2


def test_compression_hcomp_args():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        img = np.ones((20, 20))
        with fitsio.FITS(
            fn + '[compress HS 10 10; s 2.0]', 'rw', clobber=True
        ) as fits:
            fits.write(img, dither_seed=42)
        hdr = fitsio.read_header(fn, ext=1)
        for key, val in [
            ('ZTILE1', 10),
            ('ZTILE2', 10),
            ('ZQUANTIZ', 'SUBTRACTIVE_DITHER_1'),
            ('ZCMPTYPE', 'HCOMPRESS_1'),
            ('ZDITHER0', 42),
            ('ZNAME1', 'SCALE'),
            ('ZVAL1', 2.0),
            ('ZNAME2', 'SMOOTH'),
            ('ZVAL2', 1),
        ]:
            assert hdr[key] == val


def test_compression_qlevel_default():
    # Check that if not specified, qlevel defaults to 4.
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        H, W = 200, 200
        bigimg = np.random.uniform(size=(H, W))
        # Default qlevel
        with fitsio.FITS(fn, 'rw', clobber=True) as fits:
            fits.write(bigimg, compress='GZIP')
        size_def = os.stat(fn).st_size
        hdr = fitsio.read_header(fn, ext=1)
        print(hdr)
        for key, val in [
            ('ZQUANTIZ', 'SUBTRACTIVE_DITHER_1'),
            ('ZCMPTYPE', 'GZIP_1'),
        ]:
            assert hdr[key] == val
        # qlevel=0
        with fitsio.FITS(fn, 'rw', clobber=True) as fits:
            fits.write(bigimg, compress='GZIP', qlevel=0)
        size_0 = os.stat(fn).st_size
        # qlevel=4
        with fitsio.FITS(fn, 'rw', clobber=True) as fits:
            fits.write(bigimg, compress='GZIP', qlevel=4)
        size_4 = os.stat(fn).st_size
        # qlevel=16
        with fitsio.FITS(fn, 'rw', clobber=True) as fits:
            fits.write(bigimg, compress='GZIP', qlevel=16)
        size_16 = os.stat(fn).st_size
        # zero means NO COMPRESSION
        assert size_0 > size_4
        # heh, lower values mean MORE COMPRESSION
        assert size_4 < size_16
        assert size_def == size_4


def test_compression_multihdu_diskfile():
    # Check multi-HDU case with a normal file
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        img = np.ones((20, 20))
        with fitsio.FITS(fn, 'rw', clobber=True) as fits:
            # A
            fits.write(img, extname='A')
            # B
            fits.write(img, extname='B', compress='GZIP')
            # C
            fits.write(
                img,
                extname='C',
                compress='GZIP',
                qmethod='SUBTRACTIVE_DITHER_2',
            )
            # D
            fits.write(img, extname='D')
            # E
            fits.write(img, extname='E', compress='GZIP')
            # F
            fits.write(img, extname='F', compress=None)
        with fitsio.FITS(fn) as fits:
            assert len(fits) == 6
            hdrA = fits['A'].read_header()
            hdrB = fits['B'].read_header()
            hdrC = fits['C'].read_header()
            hdrD = fits['D'].read_header()
            hdrE = fits['E'].read_header()
            hdrF = fits['F'].read_header()
        # A is uncompressed
        assert 'ZCMPTYPE' not in hdrA
        # B is gzip
        assert hdrB['ZCMPTYPE'] == 'GZIP_1'
        assert hdrB['ZQUANTIZ'] == 'SUBTRACTIVE_DITHER_1'
        # C is gzip with SD2
        assert hdrC['ZCMPTYPE'] == 'GZIP_1'
        assert hdrC['ZQUANTIZ'] == 'SUBTRACTIVE_DITHER_2'
        # D is not compressed
        assert 'ZCMPTYPE' not in hdrD
        # E is GZIP again
        assert hdrE['ZCMPTYPE'] == 'GZIP_1'
        assert hdrE['ZQUANTIZ'] == 'SUBTRACTIVE_DITHER_1'
        # F is not compressed
        assert 'ZCMPTYPE' not in hdrF


def test_compression_multihdu_memfile():
    # Check multi-HDU case with a normal file
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, 'test.fits')

        img = np.ones((20, 20))
        with fitsio.FITS("mem://", 'rw', clobber=True) as fits:
            # A
            fits.write(img, extname='A')
            # B
            fits.write(img, extname='B', compress='GZIP')
            # C
            fits.write(
                img,
                extname='C',
                compress='GZIP',
                qmethod='SUBTRACTIVE_DITHER_2',
            )
            # D
            fits.write(img, extname='D')
            # E
            fits.write(img, extname='E', compress='GZIP')
            # F
            fits.write(img, extname='F', compress=None)

            data = fits.read_raw()
            with open(fn, 'wb') as f:
                f.write(data)

        with fitsio.FITS(fn) as fits:
            assert len(fits) == 6
            hdrA = fits['A'].read_header()
            hdrB = fits['B'].read_header()
            hdrC = fits['C'].read_header()
            hdrD = fits['D'].read_header()
            hdrE = fits['E'].read_header()
            hdrF = fits['F'].read_header()
        # A is uncompressed
        assert 'ZCMPTYPE' not in hdrA
        # B is gzip
        assert hdrB['ZCMPTYPE'] == 'GZIP_1'
        assert hdrB['ZQUANTIZ'] == 'SUBTRACTIVE_DITHER_1'
        # C is gzip with SD2
        assert hdrC['ZCMPTYPE'] == 'GZIP_1'
        assert hdrC['ZQUANTIZ'] == 'SUBTRACTIVE_DITHER_2'
        # D is not compressed
        assert 'ZCMPTYPE' not in hdrD
        # E is GZIP again
        assert hdrE['ZCMPTYPE'] == 'GZIP_1'
        assert hdrE['ZQUANTIZ'] == 'SUBTRACTIVE_DITHER_1'
        # F is not compressed
        assert 'ZCMPTYPE' not in hdrF



import math
import numpy as np

import pytest

from ..util import (
    _nonfinite_as_cfitsio_floating_null_value,
    cfitsio_version,
    _FLOATING_NULL_VALUE,
)

CFITSIO_VERSION = cfitsio_version(asfloat=True)
DTYPES = ['u1', 'i1', 'u2', 'i2', '<u4', 'i4', 'i8', '>f4', 'f8']
if CFITSIO_VERSION > 3.44:
    DTYPES += ["u8"]


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("with_nan", [True, False])
@pytest.mark.parametrize("hdu_is_compressed", [True, False])
def test_nonfinite_as_cfitsio_floating_null_value(
    dtype, with_nan, hdu_is_compressed
):
    data = np.arange(5 * 20, dtype=dtype).reshape(5, 20)
    if "f" in dtype and with_nan:
        data[3, 13] = np.nan
        data[1, 7] = np.inf
        data[1, 9] = -np.inf

    with _nonfinite_as_cfitsio_floating_null_value(
        data, hdu_is_compressed
    ) as (
        nan_data,
        any_nan,
    ):
        if with_nan and "f" in dtype and hdu_is_compressed:
            assert any_nan
            assert not np.any(np.isnan(nan_data))
            msk = ~np.isfinite(data)
            np.testing.assert_array_equal(data[msk], _FLOATING_NULL_VALUE)
            np.testing.assert_array_equal(data[~msk], nan_data[~msk])
        else:
            assert not any_nan
            np.testing.assert_array_equal(nan_data, data)

    if with_nan and "f" in dtype:
        np.testing.assert_array_equal(data[3, 13], np.nan)
        np.testing.assert_array_equal(data[1, 7], np.inf)
        np.testing.assert_array_equal(data[1, 9], -np.inf)


def test_cfitsio_floating_null_value_equal_inf():
    assert np.float64(np.inf) == _FLOATING_NULL_VALUE
    assert np.float32(np.inf) == _FLOATING_NULL_VALUE
    assert np.inf == _FLOATING_NULL_VALUE
    assert math.inf == _FLOATING_NULL_VALUE


def test_nonfinite_as_cfitsio_floating_null_value_with_exception():
    data = np.arange(5 * 20, dtype=np.float32).reshape(5, 20)
    data[3, 13] = np.nan
    data[1, 7] = np.inf
    data[1, 9] = -np.inf

    with pytest.raises(RuntimeError):
        with _nonfinite_as_cfitsio_floating_null_value(data, True) as (
            nan_data,
            any_nan,
        ):
            assert any_nan
            assert not np.any(np.isnan(nan_data))
            msk = ~np.isfinite(data)
            np.testing.assert_array_equal(data[msk], _FLOATING_NULL_VALUE)
            np.testing.assert_array_equal(data[~msk], nan_data[~msk])
            raise RuntimeError("Exception raised while data is modified!")

    assert data[1, 7] == np.inf
    assert data[1, 9] == -np.inf



from .base import (  # noqa
    ANY_HDU,
    BINARY_TBL,
    ASCII_TBL,
    IMAGE_HDU,
    _hdu_type_map,
)
from .image import ImageHDU  # noqa
from .table import (  # noqa
    TableHDU,
    AsciiTableHDU,
    _table_npy2fits_form,
    _npy2fits,
)



class FITSFormatError(Exception):
    """
    Format error in FITS file
    """

    def __init__(self, value):
        super(FITSFormatError, self).__init__(value)
        self.value = value

    def __str__(self):
        return str(self.value)



import pytest
import sys
import os
import tempfile
from .checks import (
    # check_header,
    compare_array,
)
import numpy as np
from ..fitslib import (
    FITS,
    read,
    write,
    RICE_1,
    SUBTRACTIVE_DITHER_1,
    GZIP_1,
    GZIP_2,
    PLIO_1,
    HCOMPRESS_1,
)
from ..util import cfitsio_is_bundled, cfitsio_version

CFITSIO_VERSION = cfitsio_version(asfloat=True)


@pytest.mark.parametrize("with_nan", [False, True])
@pytest.mark.parametrize(
    'compress',
    [
        'rice',
        'hcompress',
        'plio',
        'gzip',
        'gzip_2',
        'gzip_lossless',
        'gzip_2_lossless',
    ],
)
@pytest.mark.parametrize(
    'dtype', ['u1', 'i1', 'u2', 'i2', 'u4', 'i4', 'f4', 'f8']
)
def test_compressed_write_read(compress, dtype, with_nan):
    """
    Test writing and reading a rice compressed image
    """
    nrows = 5
    ncols = 20
    if compress in ['rice', 'hcompress'] or 'gzip' in compress:
        pass
    elif compress == 'plio':
        if dtype not in ['i1', 'i2', 'i4', 'f4', 'f8']:
            return
    else:
        raise ValueError('unexpected compress %s' % compress)

    if 'lossless' in compress:
        qlevel = None
    else:
        qlevel = 16

    seed = 1919
    rng = np.random.RandomState(seed)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        if dtype[0] == 'f':
            data = rng.normal(size=(nrows, ncols))
            if compress == 'plio':
                data = data.clip(min=0)
            data = data.astype(dtype)
        else:
            data = np.arange(
                nrows * ncols,
                dtype=dtype,
            ).reshape(nrows, ncols)

        if "f" in dtype and with_nan and compress != "plio":
            data[3, 11] = np.nan

        csend = compress.replace('_lossless', '')
        write(fname, data, compress=csend, qlevel=qlevel)
        rdata = read(fname, ext=1)

        if 'lossless' in compress or dtype[0] in ['i', 'u']:
            np.testing.assert_array_equal(
                data,
                rdata,
                err_msg="%s compressed images ('%s')" % (compress, dtype),
            )
        else:
            # lossy floating point
            np.testing.assert_allclose(
                data,
                rdata,
                rtol=0,
                atol=0.2,
                err_msg="%s compressed images ('%s')" % (compress, dtype),
            )

        with FITS(fname) as fits:
            assert fits[1].is_compressed(), "is compressed"


@pytest.mark.parametrize("with_nan", [False, True])
@pytest.mark.parametrize(
    'compress',
    [
        'rice',
        'hcompress',
        'plio',
        'gzip',
        'gzip_2',
        'gzip_lossless',
        'gzip_2_lossless',
    ],
)
@pytest.mark.parametrize(
    'dtype', ['u1', 'i1', 'u2', 'i2', 'u4', 'i4', 'f4', 'f8']
)
def test_compressed_write_read_fitsobj(compress, dtype, with_nan):
    """
    Test writing and reading a rice compressed image

    In this version, keep the fits object open
    """

    if (
        "gzip" in compress
        and dtype in ["u2", "i2", "u4", "i4"]
        and not cfitsio_is_bundled()
    ):
        pytest.xfail(
            reason=(
                "Non-bundled cfitsio libraries have a bug. "
                "See https://github.com/HEASARC/cfitsio/pull/97."
            )
        )

    nrows = 5
    ncols = 20
    if compress in ['rice', 'hcompress'] or 'gzip' in compress:
        pass
    elif compress == 'plio':
        if dtype not in ['i1', 'i2', 'i4', 'f4', 'f8']:
            return
    else:
        raise ValueError('unexpected compress %s' % compress)

    if 'lossless' in compress:
        qlevel = None
        # qlevel = 9999
    else:
        qlevel = 16

    seed = 1919
    rng = np.random.RandomState(seed)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            # note i8 not supported for compressed!

            if dtype[0] == 'f':
                data = rng.normal(size=(nrows, ncols))
                if compress == 'plio':
                    data = data.clip(min=0)
                data = data.astype(dtype)
            else:
                data = np.arange(
                    nrows * ncols,
                    dtype=dtype,
                ).reshape(nrows, ncols)

            if "f" in dtype and with_nan and compress != "plio":
                data[3, 11] = np.nan

            csend = compress.replace('_lossless', '')
            fits.write_image(data, compress=csend, qlevel=qlevel)
            rdata = fits[-1].read()

            if 'lossless' in compress or dtype[0] in ['i', 'u']:
                # for integers we have chosen a wide range of values, so
                # there will be no quantization and we expect no
                # information loss
                np.testing.assert_array_equal(
                    data,
                    rdata,
                    "%s compressed images ('%s')" % (compress, dtype),
                )
            else:
                # lossy floating point
                np.testing.assert_allclose(
                    data,
                    rdata,
                    rtol=0,
                    atol=0.2,
                    err_msg="%s compressed images ('%s')" % (compress, dtype),
                )

        with FITS(fname) as fits:
            assert fits[1].is_compressed(), "is compressed"


@pytest.mark.skipif(sys.version_info < (3, 9), reason='importlib bug in 3.8')
@pytest.mark.skipif(CFITSIO_VERSION < 3.49, reason='bug in cfitsio < 3.49')
def test_gzip_tile_compressed_read_lossless_astropy():
    """
    Test reading an image gzip compressed by astropy (fixed by cfitsio 3.49)
    """
    import importlib.resources

    ref = (
        importlib.resources.files("fitsio")
        / 'test_images'
        / 'test_gzip_compressed_image.fits.fz'
    )  # noqa
    with importlib.resources.as_file(ref) as gzip_file:
        data = read(gzip_file)

    compare_array(data, data * 0.0, "astropy lossless compressed image")


@pytest.mark.parametrize("with_nan", [False, True])
def test_compress_preserve_zeros(with_nan):
    """
    Test writing and reading gzip compressed image
    """

    zinds = [
        (1, 3),
        (2, 9),
    ]

    dtypes = ['f4', 'f8']

    seed = 2020
    rng = np.random.RandomState(seed)

    # Do not test hcompress as it doesn't support SUBTRACTIVE_DITHER_2
    for compress in ['gzip', 'gzip_2', 'rice']:
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, 'test.fits')

            with FITS(fname, 'rw') as fits:
                for dtype in dtypes:
                    data = rng.normal(size=5 * 20).reshape(5, 20).astype(dtype)
                    for zind in zinds:
                        data[zind[0], zind[1]] = 0.0
                    if with_nan:
                        data[3, 15] = np.nan

                    fits.write_image(
                        data,
                        compress=compress,
                        qlevel=16,
                        qmethod='SUBTRACTIVE_DITHER_2',
                    )
                    rdata = fits[-1].read()

                    for zind in zinds:
                        assert rdata[zind[0], zind[1]] == 0.0
                    if with_nan:
                        assert np.isnan(rdata[3, 15])


@pytest.mark.parametrize("with_nan", [False, True])
@pytest.mark.parametrize(
    'compress',
    [
        'rice',
        'hcompress',
        'plio',
    ],
)
@pytest.mark.parametrize(
    'seed_type',
    ['matched', 'unmatched', 'checksum', 'checksum_int'],
)
@pytest.mark.parametrize(
    'use_fits_object',
    [False, True],
)
@pytest.mark.parametrize(
    'dtype',
    ['f4', 'f8'],
)
def test_compressed_seed(
    compress, seed_type, use_fits_object, dtype, with_nan
):
    """
    Test writing and reading a rice compressed image
    """
    nrows = 5
    ncols = 20

    qlevel = 16

    seed = 1919
    rng = np.random.RandomState(seed)

    if seed_type == 'matched':
        # dither_seed = 9881
        dither_seed1 = 9881
        dither_seed2 = 9881
    elif seed_type == 'unmatched':
        # dither_seed = None
        dither_seed1 = 3
        dither_seed2 = 4
    elif seed_type == 'checksum':
        dither_seed1 = 'checksum'
        dither_seed2 = b'checksum'
    elif seed_type == 'checksum_int':
        dither_seed1 = -1
        # any negative means use checksum
        dither_seed2 = -3

    with tempfile.TemporaryDirectory() as tmpdir:
        fname1 = os.path.join(tmpdir, 'test1.fits')
        fname2 = os.path.join(tmpdir, 'test2.fits')

        data = rng.normal(size=(nrows, ncols))
        if compress == 'plio':
            data = data.clip(min=0)
        data = data.astype(dtype)

        if "f" in dtype and with_nan and compress != "plio":
            data[3, 11] = np.nan

        if use_fits_object:
            with FITS(fname1, 'rw') as fits1:
                fits1.write(
                    data,
                    compress=compress,
                    qlevel=qlevel,
                    # dither_seed=dither_seed,
                    dither_seed=dither_seed1,
                )
                rdata1 = fits1[-1].read()

            with FITS(fname2, 'rw') as fits2:
                fits2.write(
                    data,
                    compress=compress,
                    qlevel=qlevel,
                    # dither_seed=dither_seed,
                    dither_seed=dither_seed2,
                )
                rdata2 = fits2[-1].read()
        else:
            write(
                fname1,
                data,
                compress=compress,
                qlevel=qlevel,
                # dither_seed=dither_seed,
                dither_seed=dither_seed1,
            )
            rdata1 = read(fname1)

            write(
                fname2,
                data,
                compress=compress,
                qlevel=qlevel,
                # dither_seed=dither_seed,
                dither_seed=dither_seed2,
            )
            rdata2 = read(fname2)

        mess = "%s compressed images ('%s')" % (compress, dtype)

        if seed_type in ['checksum', 'checksum_int', 'matched']:
            np.testing.assert_array_equal(rdata1, rdata2, mess)
        else:
            with pytest.raises(AssertionError):
                np.testing.assert_array_equal(rdata1, rdata2, mess)

        if "f" in dtype and with_nan and compress != "plio":
            assert np.isnan(rdata1[3, 11])
            assert np.isnan(rdata2[3, 11])
        else:
            assert np.all(np.isfinite(rdata1))
            assert np.all(np.isfinite(rdata2))


@pytest.mark.parametrize(
    'dither_seed',
    ['blah', 10_001],
)
def test_compressed_seed_bad(dither_seed):
    """
    Test writing and reading a rice compressed image
    """
    compress = 'rice'
    dtype = 'f4'
    nrows = 5
    ncols = 20

    qlevel = 16

    seed = 1919
    rng = np.random.RandomState(seed)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        data = rng.normal(size=(nrows, ncols))
        data = data.astype(dtype)

        with pytest.raises(ValueError):
            write(
                fname,
                data,
                compress=compress,
                qlevel=qlevel,
                dither_seed=dither_seed,
            )


def test_memory_compressed_seed():
    import fitsio

    dtype = 'f4'
    nrows = 300
    ncols = 500

    seed = 1919
    rng = np.random.RandomState(seed)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname1 = os.path.join(tmpdir, 'test1.fits')
        fname2 = os.path.join(tmpdir, 'test2.fits')

        data = rng.normal(size=(nrows, ncols))
        data = data.astype(dtype)

        fitsio.write(
            fname1,
            data.copy(),
            dither_seed='checksum',
            compress='RICE',
            qlevel=1e-4,
            tile_dims=(100, 100),
            clobber=True,
        )
        hdr = fitsio.read_header(fname1, ext=1)
        dither1 = hdr['ZDITHER0']
        assert dither1 == 8269

        fits = fitsio.FITS('mem://[compress R 100,100; qz -1e-4]', 'rw')
        fits.write(data.copy(), dither_seed='checksum')
        data = fits.read_raw()
        fits.close()
        f = open(fname2, 'wb')
        f.write(data)
        f.close()
        hdr = fitsio.read_header(fname2, ext=1)
        dither2 = hdr['ZDITHER0']
        assert dither1 == dither2


def test_image_compression_inmem_subdither2():
    H, W = 100, 100
    rng = np.random.RandomState(seed=10)
    img = rng.normal(size=(H, W))
    img[40:50, :] = 0.0
    with FITS('mem://[compress G 100,100; qz 0]', 'rw') as F:
        F.write(img)
        rawdata = F.read_raw()

    with tempfile.TemporaryDirectory() as tmpdir:
        pth = os.path.join(tmpdir, 'out.fits')
        with open(pth, 'wb') as f:
            f.write(rawdata)
        im2 = read(pth)
        z = im2[40:50, :]

    minval = z.min()
    assert minval == 0


@pytest.mark.parametrize(
    "kw,val",
    [
        ("compress", RICE_1),
        ("tile_dims", (10, 2)),
        ("tile_dims", np.array([10, 2])),
        ("tile_dims", [10, 2]),
        ("qlevel", 10.0),
        ("qmethod", SUBTRACTIVE_DITHER_1),
        ("hcomp_scale", 10.0),
        ("hcomp_smooth", True),
    ],
)
@pytest.mark.parametrize("set_val_to_none", [False, True])
def test_image_compression_raises_on_python_set(kw, val, set_val_to_none):
    H, W = 100, 100
    rng = np.random.RandomState(seed=10)
    img = rng.normal(size=(H, W))
    if set_val_to_none:
        kws = {kw: None}
    else:
        kws = {kw: val}

    with FITS('mem://[compress G 100,100; qz 0]', 'rw') as F:
        with pytest.raises(ValueError):
            F.write(img, **kws)

    with FITS('mem://[compress G 100,100; qz 4.0]', 'rw') as F:
        F.write(img, dither_seed=10)


@pytest.mark.parametrize(
    "compress",
    [
        RICE_1,
        GZIP_1,
        GZIP_2,
        PLIO_1,
        HCOMPRESS_1,
    ],
)
@pytest.mark.parametrize(
    "dtype",
    [
        np.uint8,
        np.int8,
        np.uint16,
        np.int16,
        np.uint32,
        np.int32,
    ],
)
@pytest.mark.parametrize("fname", ["mem://", "test.fits"])
def test_image_compression_inmem_lossess_int(compress, dtype, fname):
    if not cfitsio_is_bundled():
        pytest.xfail(
            reason=(
                "Non-bundled cfitsio libraries have a bug. "
                "See https://github.com/HEASARC/cfitsio/pull/97 "
                "and https://github.com/HEASARC/cfitsio/pull/99."
            ),
        )
    if compress == PLIO_1 and dtype in [np.int16, np.uint32, np.int32]:
        pytest.skip(
            reason="PLIO lossless compression of int16, uint32, and "
            "int32 types is not supported by cfitsio",
        )
    rng = np.random.RandomState(seed=10)
    img = rng.normal(size=(300, 300))
    if dtype in [
        np.uint8,
        np.uint16,
        np.uint32,
    ]:
        img = np.abs(img)
    img = img.astype(dtype)
    with tempfile.TemporaryDirectory() as tmpdir:
        if "mem://" not in fname:
            fpth = os.path.join(tmpdir, fname)
        else:
            fpth = fname

        with FITS(fpth, 'rw') as F:
            F.write(img, compress=compress, qlevel=0)
            rimg = F[-1].read()
            assert rimg is not None
            assert np.array_equal(rimg, img)


def test_image_compression_inmem_lossessgzip_int_zeros():
    img = np.zeros((300, 300)).astype(np.int32)
    with FITS('mem://', 'rw') as F:
        F.write(img, compress='GZIP', qlevel=0)
        rimg = F[-1].read()
        assert rimg is not None
        assert np.array_equal(rimg, img)


def test_image_compression_inmem_lossessgzip_float():
    rng = np.random.RandomState(seed=10)
    img = rng.normal(size=(300, 300))
    with FITS('mem://', 'rw') as F:
        F.write(img, compress='GZIP', qlevel=0)
        rimg = F[-1].read()
        assert rimg is not None
        assert np.array_equal(rimg, img)


def test_image_mem_reopen_noop():
    rng = np.random.RandomState(seed=10)
    img = rng.normal(size=(300, 300))
    with FITS('mem://', 'rw') as F:
        F.write(img)
        rimg = F[-1].read()
        assert rimg is not None
        assert np.array_equal(rimg, img)
        F.reopen()
        rimg = F[-1].read()
        assert rimg is not None
        assert np.array_equal(rimg, img)

    with FITS('mem://', 'rw') as F:
        F.write(img)
        F.reopen()
        rimg = F[-1].read()
        assert rimg is not None
        assert np.array_equal(rimg, img)


@pytest.mark.parametrize("nan_value", [np.nan, np.inf, -np.inf])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize(
    "fname",
    [
        "test.fits",
        "mem://",
    ],
)
def test_image_compression_nulls(fname, dtype, nan_value):
    data = np.arange(36).reshape((6, 6)).astype(dtype)
    data[1, 1] = nan_value

    # everything comes back as nan
    if nan_value is not np.nan:
        msk = ~np.isfinite(data)
        cdata = data.copy()
        cdata[msk] = np.nan
    else:
        cdata = data

    with tempfile.TemporaryDirectory() as tmpdir:
        if "mem://" not in fname:
            fpth = os.path.join(tmpdir, fname)
        else:
            fpth = fname

        with FITS(fpth, "rw") as fits:
            fits.write(
                data,
                compress='RICE_1',
                tile_dims=(3, 3),
                dither_seed=10,
                qlevel=2,
            )
            read_data = fits[1].read()

            np.testing.assert_allclose(
                read_data,
                cdata,
            )

        if "mem://" not in fpth:
            with FITS(fpth, "r") as fits:
                read_data = fits[1].read()
                np.testing.assert_allclose(
                    read_data,
                    cdata,
                )


if __name__ == '__main__':
    test_compressed_seed(
        compress='rice',
        match_seed=False,
        use_fits_object=True,
        dtype='f4',
    )



import os
import tempfile
import warnings
import numpy as np

import pytest

from .makedata import make_data, lorem_ipsum
from .checks import check_header, compare_headerlist_header
from ..fitslib import FITS, read_header, write
from ..header import FITSHDR
from ..hdu.base import INVALID_HDR_CHARS
from ..util import cfitsio_version

CFITSIO_VERSION = cfitsio_version(asfloat=True)


def test_free_form_string():
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')
        with open(fname, 'w') as f:
            s = (
                "SIMPLE  =                    T / Standard FITS                                  "  # noqa
                + "BITPIX  =                   16 / number of bits per data pixel                  "  # noqa
                + "NAXIS   =                    0 / number of data axes                            "  # noqa
                + "EXTEND  =                    T / File contains extensions                       "  # noqa
                + "PHOTREF =   'previous MegaCam' / Source: cum.photcat                            "  # noqa
                + "EXTRA   =                    7 / need another line following PHOTREF            "  # noqa
                + "END                                                                             "  # noqa
            )
            f.write(s + ' ' * (2880 - len(s)))
        hdr = read_header(fname)
        assert hdr['PHOTREF'] == 'previous MegaCam'


def test_add_delete_and_update_records():
    # Build a FITSHDR from a few records (no need to write on disk)
    # Record names have to be in upper case to match with FITSHDR.add_record
    recs = [
        {'name': "First_record".upper(), 'value': 1, 'comment': "number 1"},
        {'name': "Second_record".upper(), 'value': "2"},
        {'name': "Third_record".upper(), 'value': "3"},
        {'name': "Last_record".upper(), 'value': 4, 'comment': "number 4"},
    ]
    hdr = FITSHDR(recs)

    # Add a new record
    hdr.add_record({'name': 'New_record'.upper(), 'value': 5})

    # Delete number 2 and 4
    hdr.delete('Second_record'.upper())
    hdr.delete('Last_record'.upper())

    # Update records : first and new one
    hdr['First_record'] = 11
    hdr['New_record'] = 3

    # Do some checks : len and get value/comment
    assert len(hdr) == 3
    assert hdr['First_record'] == 11
    assert hdr['New_record'] == 3
    assert hdr['Third_record'] == '3'
    assert hdr.get_comment('First_record') == 'number 1'
    assert not hdr.get_comment('New_record')


def testHeaderCommentPreserved():
    """
    Test that the comment is preserved after resetting the value
    """

    l1 = 'KEY1    =                   77 / My comment1'
    l2 = 'KEY2    =                   88 / My comment2'
    hdr = FITSHDR()
    hdr.add_record(l1)
    hdr.add_record(l2)

    hdr['key1'] = 99
    assert hdr.get_comment('key1') == 'My comment1', 'comment not preserved'


def test_header_write_read():
    """
    Test a basic header write and read

    Note the other read/write tests also are checking header writing with
    a list of dicts
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.zeros(10)
            header = {
                'x': 35,
                'y': 88.215,
                'eval': 1.384123233e43,
                'empty': '',
                'funky': '35-8',  # test old bug when strings look
                # like expressions
                'name': 'J. Smith',
                'what': '89113e6',  # test bug where converted to float
                'und': None,
                'binop': '25-3',  # test string with binary operation in it
                'unders': '1_000_000',  # test string with underscore
                'longs': lorem_ipsum,
            }
            if CFITSIO_VERSION > 4.02:
                # force hierarch + continue
                header["long_keyword_name"] = lorem_ipsum

            fits.write_image(data, header=header)

            rh = fits[0].read_header()
            check_header(header, rh)

        with FITS(fname) as fits:
            rh = fits[0].read_header()
            check_header(header, rh)


def test_header_delete():
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.zeros(10)
            header1 = {'SCARD': 'one', 'ICARD': 1, 'FCARD': 1.0, 'LCARD': True}
            fits.write_image(data, header=header1)
            rh = fits[0].read_header()
            check_header(header1, rh)

            fits[0].delete_key("SCARD")
            del header1["SCARD"]
            rh = fits[0].read_header()
            check_header(header1, rh)

            fits[0].delete_keys(["ICARD", "FCARD"])
            del header1["ICARD"]
            del header1["FCARD"]
            rh = fits[0].read_header()
            check_header(header1, rh)

        with FITS(fname) as fits:
            rh = fits[0].read_header()
            check_header(header1, rh)


def test_header_update():
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.zeros(10)
            header1 = {'SCARD': 'one', 'ICARD': 1, 'FCARD': 1.0, 'LCARD': True}
            header2 = {
                'SCARD': 'two',
                'ICARD': 2,
                'FCARD': 2.0,
                'LCARD': False,
                'SNEW': 'two',
                'INEW': 2,
                'FNEW': 2.0,
                'LNEW': False,
            }
            fits.write_image(data, header=header1)
            rh = fits[0].read_header()
            check_header(header1, rh)

            # Update header
            fits[0].write_keys(header2)

        with FITS(fname) as fits:
            rh = fits[0].read_header()
            check_header(header2, rh)


def test_read_header_case():
    """
    Test read_header with and without case sensitivity

    The reason we need a special test for this is because
    the read_header code is optimized for speed and has
    a different code path
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.zeros(10)
            adata = make_data()
            fits.write_image(data, header=adata['keys'], extname='First')
            fits.write_image(data, header=adata['keys'], extname='second')

        cases = [
            ('First', True),
            ('FIRST', False),
            ('second', True),
            ('seConD', False),
        ]
        for ext, ci in cases:
            h = read_header(fname, ext=ext, case_sensitive=ci)
            compare_headerlist_header(adata['keys'], h)


def test_blank_key_comments():
    """
    test a few different comments
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            records = [
                # empty should return empty
                {'name': None, 'value': '', 'comment': ''},
                # this will also return empty
                {'name': None, 'value': '', 'comment': ' '},
                # this will return exactly
                {'name': None, 'value': '', 'comment': ' h'},
                # this will return exactly
                {'name': None, 'value': '', 'comment': '--- test comment ---'},
            ]
            header = FITSHDR(records)

            fits.write(None, header=header)

            rh = fits[0].read_header()

            rrecords = rh.records()

            for i, ri in ((0, 6), (1, 7), (2, 8)):
                rec = records[i]
                rrec = rrecords[ri]

                assert rec['name'] is None, 'checking name is None'

                comment = rec['comment']
                rcomment = rrec['comment']
                if '' == comment.strip():
                    comment = ''

                assert comment == rcomment, "check empty key comment"


def test_blank_key_comments_from_cards():
    """
    test a few different comments
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            records = [
                '                                                                                ',  # noqa
                '         --- testing comment ---                                                ',  # noqa
                '        --- testing comment ---                                                 ',  # noqa
                "COMMENT testing                                                                 ",  # noqa
            ]
            header = FITSHDR(records)

            fits.write(None, header=header)

            rh = fits[0].read_header()

            rrecords = rh.records()

            assert rrecords[6]['name'] is None, 'checking name is None'
            assert rrecords[6]['comment'] == '', 'check empty key comment'
            assert rrecords[7]['name'] is None, 'checking name is None'
            assert rrecords[7]['comment'] == ' --- testing comment ---', (
                "check empty key comment"
            )
            assert rrecords[8]['name'] is None, 'checking name is None'
            assert rrecords[8]['comment'] == '--- testing comment ---', (
                "check empty key comment"
            )
            assert rrecords[9]['name'] == 'COMMENT', 'checking name is COMMENT'
            assert rrecords[9]['comment'] == 'testing', "check comment"


def test_header_from_cards():
    """
    test generating a header from cards, writing it out and getting
    back what we put in
    """
    hdr_from_cards = FITSHDR(
        [
            "IVAL    =                   35 / integer value                                  ",  # noqa
            "SHORTS  = 'hello world'                                                         ",  # noqa
            "UND     =                                                                       ",  # noqa
            "LONGS   = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiu&'",  # noqa
            "CONTINUE  'smod tempor incididunt ut labore et dolore magna aliqua'             ",  # noqa
            "DBL     =                 1.25                                                  ",  # noqa
        ]
    )
    header = [
        {'name': 'ival', 'value': 35, 'comment': 'integer value'},
        {'name': 'shorts', 'value': 'hello world'},
        {'name': 'und', 'value': None},
        {'name': 'longs', 'value': lorem_ipsum},
        {'name': 'dbl', 'value': 1.25},
    ]

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.zeros(10)
            fits.write_image(data, header=hdr_from_cards)

            rh = fits[0].read_header()
            compare_headerlist_header(header, rh)

        with FITS(fname) as fits:
            rh = fits[0].read_header()
            compare_headerlist_header(header, rh)


def test_bad_header_write_raises():
    """
    Test that an invalid header raises.
    """

    for c in INVALID_HDR_CHARS:
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, 'test.fits')
            try:
                hdr = {'bla%sg' % c: 3}
                data = np.zeros(10)

                write(fname, data, header=hdr, clobber=True)
            except Exception as e:
                assert "header key 'BLA%sG' has" % c in str(e)


def test_header_template():
    """
    test adding bunch of cards from a split template
    """

    header_template = """SIMPLE  =                    T /
BITPIX  =                    8 / bits per data value
NAXIS   =                    0 / number of axes
EXTEND  =                    T / Extensions are permitted
ORIGIN  = 'LSST DM Header Service'/ FITS file originator

     ---- Date, night and basic image information ----
DATE    =                      / Creation Date and Time of File
DATE-OBS=                      / Date of the observation (image acquisition)
DATE-BEG=                      / Time at the start of integration
DATE-END=                      / end date of the observation
MJD     =                      / Modified Julian Date that the file was written
MJD-OBS =                      / Modified Julian Date of observation
MJD-BEG =                      / Modified Julian Date derived from DATE-BEG
MJD-END =                      / Modified Julian Date derived from DATE-END
OBSID   =                      / ImageName from Camera StartIntergration
GROUPID =                      / imageSequenceName from StartIntergration
OBSTYPE =                      / BIAS, DARK, FLAT, OBJECT
BUNIT   = 'adu     '           / Brightness units for pixel array

     ---- Telescope info, location, observer ----
TELESCOP= 'LSST AuxTelescope'  / Telescope name
INSTRUME= 'LATISS'             / Instrument used to obtain these data
OBSERVER= 'LSST'               / Observer name(s)
OBS-LONG=           -70.749417 / [deg] Observatory east longitude
OBS-LAT =           -30.244639 / [deg] Observatory latitude
OBS-ELEV=               2663.0 / [m] Observatory elevation
OBSGEO-X=           1818938.94 / [m] X-axis Geocentric coordinate
OBSGEO-Y=          -5208470.95 / [m] Y-axis Geocentric coordinate
OBSGEO-Z=          -3195172.08 / [m] Z-axis Geocentric coordinate

    ---- Pointing info, etc. ----

DECTEL  =                      / Telescope DEC of observation
ROTPATEL=                      / Telescope Rotation
ROTCOORD= 'sky'                / Telescope Rotation Coordinates
RA      =                      / RA of Target
DEC     =                      / DEC of Target
ROTPA   =                      / Rotation angle relative to the sky (deg)
HASTART =                      / [HH:MM:SS] Telescope hour angle at start
ELSTART =                      / [deg] Telescope zenith distance at start
AZSTART =                      / [deg] Telescope azimuth angle at start
AMSTART =                      / Airmass at start
HAEND   =                      / [HH:MM:SS] Telescope hour angle at end
ELEND   =                      / [deg] Telescope zenith distance at end
AZEND   =                      / [deg] Telescope azimuth angle at end
AMEND   =                      / Airmass at end

    ---- Image-identifying used to build OBS-ID ----
TELCODE = 'AT'                 / The code for the telecope
CONTRLLR=                      / The controller (e.g. O for OCS, C for CCS)
DAYOBS  =                      / The observation day as defined by image name
SEQNUM  =                      / The sequence number from the image name
GROUPID =                      /

    ---- Information from Camera
CCD_MANU= 'ITL'                / CCD Manufacturer
CCD_TYPE= '3800C'              / CCD Model Number
CCD_SERN= '20304'              / Manufacturers? CCD Serial Number
LSST_NUM= 'ITL-3800C-098'      / LSST Assigned CCD Number
SEQCKSUM=                      / Checksum of Sequencer
SEQNAME =                      / SequenceName from Camera StartIntergration
REBNAME =                      / Name of the REB
CONTNUM =                      / CCD Controller (WREB) Serial Number
IMAGETAG=                      / DAQ Image id
TEMP_SET=                      / Temperature set point (deg C)
CCDTEMP =                      / Measured temperature (deg C)

    ---- Geometry from Camera ----
DETSIZE =                      / Size of sensor
OVERH   =                      / Over-scan pixels
OVERV   =                      / Vert-overscan pix
PREH    =                      / Pre-scan pixels

    ---- Filter/grating information ----
FILTER  =                      / Name of the filter
FILTPOS =                      / Filter position
GRATING =                      / Name of the second disperser
GRATPOS =                      / disperser position
LINSPOS =                      / Linear Stage

    ---- Exposure-related information ----
EXPTIME =                      / Exposure time in seconds
SHUTTIME=                      / Shutter exposure time in seconds
DARKTIME=                      / Dark time in seconds

    ---- Header information ----
FILENAME=                      / Original file name
HEADVER =                      / Version of header

    ---- Checksums ----
CHECKSUM=                      / checksum for the current HDU
DATASUM =                      / checksum of the data records\n"""

    lines = header_template.splitlines()
    hdr = FITSHDR()
    for line in lines:
        hdr.add_record(line)


def test_corrupt_continue():
    """
    test with corrupt continue, just make sure it doesn't crash
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')
        with warnings.catch_warnings(record=True) as _:
            hdr_from_cards = FITSHDR(
                [
                    "IVAL    =                   35 / integer value                                  ",  # noqa
                    "SHORTS  = 'hello world'                                                         ",  # noqa
                    "CONTINUE= '        '           /   '&' / Current observing orogram              ",  # noqa
                    "UND     =                                                                       ",  # noqa
                    "DBL     =                 1.25                                                  ",  # noqa
                ]
            )

            with FITS(fname, 'rw') as fits:
                fits.write(None, header=hdr_from_cards)

            read_header(fname)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')
        with warnings.catch_warnings(record=True) as _:
            hdr_from_cards = FITSHDR(
                [
                    "IVAL    =                   35 / integer value                                  ",  # noqa
                    "SHORTS  = 'hello world'                                                         ",  # noqa
                    "PROGRAM = 'Setting the Scale: Determining the Absolute Mass Normalization and &'",  # noqa
                    "CONTINUE  'Scaling Relations for Clusters at z~0.1&'                            ",  # noqa
                    "CONTINUE  '&' / Current observing orogram                                       ",  # noqa
                    "UND     =                                                                       ",  # noqa
                    "DBL     =                 1.25                                                  ",  # noqa
                ]
            )

            with FITS(fname, 'rw') as fits:
                fits.write(None, header=hdr_from_cards)

            read_header(fname)


def record_exists(header_records, key, value):
    for rec in header_records:
        if rec['name'] == key and rec['value'] == value:
            return True

    return False


def test_read_comment_history():
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.arange(100).reshape(10, 10)
            fits.create_image_hdu(data)
            hdu = fits[-1]
            hdu.write_comment('A COMMENT 1')
            hdu.write_comment('A COMMENT 2')
            hdu.write_history('SOME HISTORY 1')
            hdu.write_history('SOME HISTORY 2')
            fits.close()

        with FITS(fname, 'r') as fits:
            hdu = fits[-1]
            header = hdu.read_header()
            records = header.records()
            assert record_exists(records, 'COMMENT', 'A COMMENT 1')
            assert record_exists(records, 'COMMENT', 'A COMMENT 2')
            assert record_exists(records, 'HISTORY', 'SOME HISTORY 1')
            assert record_exists(records, 'HISTORY', 'SOME HISTORY 2')


def test_write_key_dict():
    """
    test that write_key works using a standard key dict
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')
        with FITS(fname, 'rw') as fits:
            im = np.zeros((10, 10), dtype='i2')
            fits.write(im)

            keydict = {
                'name': 'test',
                'value': 35,
                'comment': 'keydict test',
            }
            fits[-1].write_key(**keydict)

            h = fits[-1].read_header()

            assert h['test'] == keydict['value']
            assert h.get_comment('test') == keydict['comment']


@pytest.mark.parametrize("fname", ["test.fits", "mem://"])
def test_header_update_compressed_image_to_table(fname):
    data = np.arange(10).reshape(5, 2).astype(np.float32)

    fname = "test.fits"
    with tempfile.TemporaryDirectory() as tmpdir:
        if "mem://" not in fname:
            fpth = os.path.join(tmpdir, fname)
        else:
            fpth = fname

        with FITS(fpth, "rw") as fits:
            fits.write(data, compress="RICE", qlevel=1, dither_seed=10)
            hdr = fits[1].read_header()

            info_before = fits[1].get_info()
            for key in hdr.keys():
                if key.startswith("Z"):
                    fits[1].delete_key(key)
            for i in range(1000):
                fits[1].write_key("test" + str(i), "blah")
            fits[1].delete_key("test0")

            fits.update_hdu_list()
            info_after = fits[1].get_info()

            assert info_after != info_before
            assert fits[1].get_exttype() == "BINARY_TBL"


if __name__ == '__main__':
    test_header_write_read()



import sys
import numpy as np
from functools import lru_cache

from ..util import cfitsio_version

CFITSIO_VERSION = cfitsio_version(asfloat=True)

lorem_ipsum = (
    'Lorem ipsum dolor sit amet, consectetur adipiscing '
    'elit, sed do eiusmod tempor incididunt ut labore '
    'et dolore magna aliqua'
)


@lru_cache(maxsize=1)
def make_data():
    nvec = 2
    ashape = (21, 21)
    Sdtype = 'S6'
    Udtype = 'U6'

    # all currently available types, scalar, 1-d and 2-d array columns
    dtype = [
        ('u1scalar', 'u1'),
        ('i1scalar', 'i1'),
        ('b1scalar', '?'),
        ('u2scalar', 'u2'),
        ('i2scalar', 'i2'),
        ('u4scalar', 'u4'),
        ('i4scalar', '<i4'),  # mix the byte orders a bit, test swapping
        ('i8scalar', 'i8'),
        ('f4scalar', 'f4'),
        ('f8scalar', '>f8'),
        ('c8scalar', 'c8'),  # complex, two 32-bit
        ('c16scalar', 'c16'),  # complex, two 32-bit
        ('u1vec', 'u1', nvec),
        ('i1vec', 'i1', nvec),
        ('b1vec', '?', nvec),
        ('u2vec', 'u2', nvec),
        ('i2vec', 'i2', nvec),
        ('u4vec', 'u4', nvec),
        ('i4vec', 'i4', nvec),
        ('i8vec', 'i8', nvec),
        ('f4vec', 'f4', nvec),
        ('f8vec', 'f8', nvec),
        ('c8vec', 'c8', nvec),
        ('c16vec', 'c16', nvec),
        ('u1arr', 'u1', ashape),
        ('i1arr', 'i1', ashape),
        ('b1arr', '?', ashape),
        ('u2arr', 'u2', ashape),
        ('i2arr', 'i2', ashape),
        ('u4arr', 'u4', ashape),
        ('i4arr', 'i4', ashape),
        ('i8arr', 'i8', ashape),
        ('f4arr', 'f4', ashape),
        ('f8arr', 'f8', ashape),
        ('c8arr', 'c8', ashape),
        ('c16arr', 'c16', ashape),
        # special case of (1,)
        ('f8arr_dim1', 'f8', (1,)),
        ('Sscalar', Sdtype),
        ('Svec', Sdtype, nvec),
        ('Sarr', Sdtype, ashape),
    ]

    if CFITSIO_VERSION > 4:
        dtype += [
            ('u8scalar', 'u8'),
            ('u8vec', 'u8', nvec),
            ('u8arr', 'u8', ashape),
        ]

    # cfitsio 3 or earlier does not
    # handle non-space padded strings
    # properly
    if CFITSIO_VERSION >= 4:
        dtype += [
            ('Sscalar_nopad', Sdtype),
            ('Svec_nopad', Sdtype, nvec),
            ('Sarr_nopad', Sdtype, ashape),
        ]

    if sys.version_info > (3, 0, 0):
        dtype += [
            ('Uscalar', Udtype),
            ('Uvec', Udtype, nvec),
            ('Uarr', Udtype, ashape),
        ]

        # cfitsio 3 or earlier does not
        # handle non-space padded strings
        # properly
        if CFITSIO_VERSION >= 4:
            dtype += [
                ('Uscalar_nopad', Udtype),
                ('Uvec_nopad', Udtype, nvec),
                ('Uarr_nopad', Udtype, ashape),
            ]

    dtype2 = [
        ('index', 'i4'),
        ('x', 'f8'),
        ('y', 'f8'),
    ]

    nrows = 4
    data = np.zeros(nrows, dtype=dtype)

    dtypes = [
        'u1',
        'i1',
        'u2',
        'i2',
        'u4',
        'i4',
        'i8',
        'f4',
        'f8',
        'c8',
        'c16',
    ]
    if CFITSIO_VERSION > 4:
        dtypes += ["u8"]

    for t in dtypes:
        if t in ['c8', 'c16']:
            data[t + 'scalar'] = [
                complex(i + 1, (i + 1) * 2) for i in range(nrows)
            ]
            vname = t + 'vec'
            for row in range(nrows):
                for i in range(nvec):
                    index = (row + 1) * (i + 1)
                    data[vname][row, i] = complex(index, index * 2)
            aname = t + 'arr'
            for row in range(nrows):
                for i in range(ashape[0]):
                    for j in range(ashape[1]):
                        index = (row + 1) * (i + 1) * (j + 1)
                        data[aname][row, i, j] = complex(index, index * 2)

        else:
            data[t + 'scalar'] = 1 + np.arange(nrows, dtype=t)
            data[t + 'vec'] = 1 + np.arange(
                nrows * nvec,
                dtype=t,
            ).reshape(nrows, nvec)
            arr = 1 + np.arange(nrows * ashape[0] * ashape[1], dtype=t)
            data[t + 'arr'] = arr.reshape(nrows, ashape[0], ashape[1])

    for t in ['b1']:
        data[t + 'scalar'] = (np.arange(nrows) % 2 == 0).astype('?')
        data[t + 'vec'] = (
            (np.arange(nrows * nvec) % 2 == 0).astype('?').reshape(nrows, nvec)
        )

        arr = (np.arange(nrows * ashape[0] * ashape[1]) % 2 == 0).astype('?')
        data[t + 'arr'] = arr.reshape(nrows, ashape[0], ashape[1])

    # strings get padded when written to the fits file.  And the way I do
    # the read, I read all bytes (ala mrdfits) so the spaces are preserved.
    #
    # so we need to pad out the strings with blanks so we can compare

    data['Sscalar'] = ['%-6s' % s for s in ['hello', 'world', 'good', 'bye']]
    data['Svec'][:, 0] = '%-6s' % 'hello'
    data['Svec'][:, 1] = '%-6s' % 'world'

    s = 1 + np.arange(nrows * ashape[0] * ashape[1])
    s = ['%-6s' % el for el in s]
    data['Sarr'] = np.array(s).reshape(nrows, ashape[0], ashape[1])

    # cfitsio 3 or earlier does not
    # handle non-space padded strings
    # properly
    if CFITSIO_VERSION >= 4:
        data['Sscalar_nopad'] = ['hello', 'world', 'good', 'bye']
        data['Svec_nopad'][:, 0] = 'hello'
        data['Svec_nopad'][:, 1] = 'world'

        s = 1 + np.arange(nrows * ashape[0] * ashape[1])
        s = ['%s' % el for el in s]
        data['Sarr_nopad'] = np.array(s).reshape(nrows, ashape[0], ashape[1])

    if sys.version_info >= (3, 0, 0):
        data['Uscalar'] = [
            '%-6s' % s for s in ['hello', 'world', 'good', 'bye']
        ]
        data['Uvec'][:, 0] = '%-6s' % 'hello'
        data['Uvec'][:, 1] = '%-6s' % 'world'

        s = 1 + np.arange(nrows * ashape[0] * ashape[1])
        s = ['%-6s' % el for el in s]
        data['Uarr'] = np.array(s).reshape(nrows, ashape[0], ashape[1])

        # cfitsio 3 or earlier does not
        # handle non-space padded strings
        # properly
        if CFITSIO_VERSION >= 4:
            data['Uscalar_nopad'] = ['hello', 'world', 'good', 'bye']
            data['Uvec_nopad'][:, 0] = 'hello'
            data['Uvec_nopad'][:, 1] = 'world'

            s = 1 + np.arange(nrows * ashape[0] * ashape[1])
            s = ['%s' % el for el in s]
            data['Uarr_nopad'] = np.array(s).reshape(
                nrows,
                ashape[0],
                ashape[1],
            )

    # use a dict list so we can have comments
    # for long key we used the largest possible

    keys = [
        {'name': 'test1', 'value': 35},
        {'name': 'empty', 'value': ''},
        {'name': 'long_keyword_name', 'value': 'stuff'},
        {
            'name': 'test2',
            'value': 'stuff',
            'comment': 'this is a string keyword',
        },
        {
            'name': 'dbl',
            'value': 23.299843,
            'comment': "this is a double keyword",
        },
        {
            'name': 'edbl',
            'value': 1.384123233e43,
            'comment': "double keyword with exponent",
        },
        {
            'name': 'lng',
            'value': 2**63 - 1,
            'comment': 'this is a long keyword',
        },
        {'name': 'lngstr', 'value': lorem_ipsum, 'comment': 'long string'},
    ]

    # a second extension using the convenience function
    nrows2 = 10
    data2 = np.zeros(nrows2, dtype=dtype2)
    data2['index'] = np.arange(nrows2, dtype='i4')
    data2['x'] = np.arange(nrows2, dtype='f8')
    data2['y'] = np.arange(nrows2, dtype='f8')

    #
    # ascii table
    #

    nvec = 2
    ashape = (2, 3)
    Sdtype = 'S6'
    Udtype = 'U6'

    # we support writing i2, i4, i8, f4 f8, but when reading cfitsio always
    # reports their types as i4 and f8, so can't really use i8 and we are
    # forced to read all floats as f8 precision

    adtype = [
        ('i2scalar', 'i2'),
        ('i4scalar', 'i4'),
        #  ('i8scalar', 'i8'),
        ('f4scalar', 'f4'),
        ('f8scalar', 'f8'),
        ('Sscalar', Sdtype),
    ]
    if sys.version_info >= (3, 0, 0):
        adtype += [('Uscalar', Udtype)]

    nrows = 4
    try:
        tdt = np.dtype(adtype, align=True)
    except TypeError:  # older numpy may not understand `align` argument
        tdt = np.dtype(adtype)
    adata = np.zeros(nrows, dtype=tdt)

    adata['i2scalar'][:] = -32222 + np.arange(nrows, dtype='i2')
    adata['i4scalar'][:] = -1353423423 + np.arange(nrows, dtype='i4')
    adata['f4scalar'][:] = (
        -2.55555555555555555555555e35 + np.arange(nrows, dtype='f4') * 1.0e35
    )
    adata['f8scalar'][:] = (
        -2.55555555555555555555555e110 + np.arange(nrows, dtype='f8') * 1.0e110
    )
    adata['Sscalar'] = ['hello', 'world', 'good', 'bye']

    if sys.version_info >= (3, 0, 0):
        adata['Uscalar'] = ['hello', 'world', 'good', 'bye']

    ascii_data = adata

    #
    # for variable length columns
    #

    # all currently available types, scalar, 1-d and 2-d array columns
    dtype = [
        ('u1scalar', 'u1'),
        ('u1obj', 'O'),
        ('i1scalar', 'i1'),
        ('i1obj', 'O'),
        ('u2scalar', 'u2'),
        ('u2obj', 'O'),
        ('i2scalar', 'i2'),
        ('i2obj', 'O'),
        ('u4scalar', 'u4'),
        ('u4obj', 'O'),
        ('i4scalar', '<i4'),  # mix the byte orders a bit, test swapping
        ('i4obj', 'O'),
        ('i8scalar', 'i8'),
        ('i8obj', 'O'),
        ('f4scalar', 'f4'),
        ('f4obj', 'O'),
        ('f8scalar', '>f8'),
        ('f8obj', 'O'),
        ('u1vec', 'u1', nvec),
        ('i1vec', 'i1', nvec),
        ('u2vec', 'u2', nvec),
        ('i2vec', 'i2', nvec),
        ('u4vec', 'u4', nvec),
        ('i4vec', 'i4', nvec),
        ('i8vec', 'i8', nvec),
        ('f4vec', 'f4', nvec),
        ('f8vec', 'f8', nvec),
        ('u1arr', 'u1', ashape),
        ('i1arr', 'i1', ashape),
        ('u2arr', 'u2', ashape),
        ('i2arr', 'i2', ashape),
        ('u4arr', 'u4', ashape),
        ('i4arr', 'i4', ashape),
        ('i8arr', 'i8', ashape),
        ('f4arr', 'f4', ashape),
        ('f8arr', 'f8', ashape),
        # special case of (1,)
        ('f8arr_dim1', 'f8', (1,)),
        ('Sscalar', Sdtype),
        ('Sobj', 'O'),
        ('Svec', Sdtype, nvec),
        ('Sarr', Sdtype, ashape),
    ]
    if CFITSIO_VERSION > 4:
        dtype += [
            ('u8vec', 'u8', nvec),
            ('u8arr', 'u8', ashape),
            ('u8scalar', 'i8'),
            ('u8obj', 'O'),
        ]

    if sys.version_info > (3, 0, 0):
        dtype += [
            ('Uscalar', Udtype),
            ('Uvec', Udtype, nvec),
            ('Uarr', Udtype, ashape),
        ]

    nrows = 4
    vardata = np.zeros(nrows, dtype=dtype)

    _dtypes = ['u1', 'i1', 'u2', 'i2', 'u4', 'i4', 'i8', 'f4', 'f8']
    if CFITSIO_VERSION > 4:
        _dtypes += ["u8"]
    for t in _dtypes:
        vardata[t + 'scalar'] = 1 + np.arange(nrows, dtype=t)
        vardata[t + 'vec'] = 1 + np.arange(nrows * nvec, dtype=t).reshape(
            nrows,
            nvec,
        )
        arr = 1 + np.arange(nrows * ashape[0] * ashape[1], dtype=t)
        vardata[t + 'arr'] = arr.reshape(nrows, ashape[0], ashape[1])

        for i in range(nrows):
            vardata[t + 'obj'][i] = vardata[t + 'vec'][i]

    # strings get padded when written to the fits file.  And the way I do
    # the read, I real all bytes (ala mrdfits) so the spaces are preserved.
    #
    # so for comparisons, we need to pad out the strings with blanks so we
    # can compare

    vardata['Sscalar'] = [
        '%-6s' % s for s in ['hello', 'world', 'good', 'bye']
    ]
    vardata['Svec'][:, 0] = '%-6s' % 'hello'
    vardata['Svec'][:, 1] = '%-6s' % 'world'

    s = 1 + np.arange(nrows * ashape[0] * ashape[1])
    s = ['%-6s' % el for el in s]
    vardata['Sarr'] = np.array(s).reshape(nrows, ashape[0], ashape[1])

    if sys.version_info >= (3, 0, 0):
        vardata['Uscalar'] = [
            '%-6s' % s for s in ['hello', 'world', 'good', 'bye']
        ]
        vardata['Uvec'][:, 0] = '%-6s' % 'hello'
        vardata['Uvec'][:, 1] = '%-6s' % 'world'

        s = 1 + np.arange(nrows * ashape[0] * ashape[1])
        s = ['%-6s' % el for el in s]
        vardata['Uarr'] = np.array(s).reshape(nrows, ashape[0], ashape[1])

    for i in range(nrows):
        vardata['Sobj'][i] = vardata['Sscalar'][i].rstrip()

    #
    # for bitcol columns
    #
    nvec = 2
    ashape = (21, 21)

    dtype = [('b1vec', '?', nvec), ('b1arr', '?', ashape)]

    nrows = 4
    bdata = np.zeros(nrows, dtype=dtype)

    for t in ['b1']:
        bdata[t + 'vec'] = (
            (np.arange(nrows * nvec) % 2 == 0)
            .astype('?')
            .reshape(
                nrows,
                nvec,
            )
        )
        arr = (np.arange(nrows * ashape[0] * ashape[1]) % 2 == 0).astype('?')
        bdata[t + 'arr'] = arr.reshape(nrows, ashape[0], ashape[1])

    return {
        'data': data,
        'keys': keys,
        'data2': data2,
        'ascii_data': ascii_data,
        'vardata': vardata,
        'bdata': bdata,
    }



import pytest
import numpy as np
import os
import tempfile
from .checks import (
    compare_names,
    compare_array,
    compare_array_tol,
    compare_object_array,
    compare_rec,
    compare_headerlist_header,
    compare_rec_with_var,
    compare_rec_subrows,
)
from .makedata import make_data
from ..fitslib import FITS, write, read
from .. import util
from .. import cfitsio_has_bzip2_support

CFITSIO_VERSION = util.cfitsio_version(asfloat=True)
DTYPES = ['u1', 'i1', 'u2', 'i2', '<u4', 'i4', 'i8', '>f4', 'f8']
if CFITSIO_VERSION > 4:
    DTYPES += ["u8"]


def test_table_read_write():
    adata = make_data()

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(
                adata['data'], header=adata['keys'], extname='mytable'
            )

            d = fits[1].read()
            compare_rec(adata['data'], d, "table read/write")

            h = fits[1].read_header()
            compare_headerlist_header(adata['keys'], h)

        # see if our convenience functions are working
        write(
            fname,
            adata['data2'],
            extname="newext",
            header={'ra': 335.2, 'dec': -25.2},
        )
        d = read(fname, ext='newext')
        compare_rec(adata['data2'], d, "table data2")

        # now test read_column
        with FITS(fname) as fits:
            for f in adata['data'].dtype.names:
                d = fits[1].read_column(f)
                compare_array(
                    adata['data'][f], d, "table 1 single field read '%s'" % f
                )

            for f in adata['data2'].dtype.names:
                d = fits['newext'].read_column(f)
                compare_array(
                    adata['data2'][f], d, "table 2 single field read '%s'" % f
                )

            # now list of columns
            for cols in [
                ['u2scalar', 'f4vec', 'Sarr'],
                ['f8scalar', 'u2arr', 'Sscalar'],
            ]:
                d = fits[1].read(columns=cols)
                for f in d.dtype.names:
                    compare_array(
                        adata['data'][f][:], d[f], "test column list %s" % f
                    )

                for rows in [[1, 3], [3, 1], [2, 2, 1]]:
                    d = fits[1].read(columns=cols, rows=rows)
                    for col in d.dtype.names:
                        compare_array(
                            adata['data'][col][rows],
                            d[col],
                            "test column list %s row subset" % col,
                        )
                    for col in cols:
                        d = fits[1].read_column(col, rows=rows)
                        compare_array(
                            adata['data'][col][rows],
                            d,
                            "test column list %s row subset" % col,
                        )


@pytest.mark.parametrize('nvec', [2, 1])
def test_table_read_write_vec1(nvec):
    """
    ensure the data for vec length 1 gets round-tripped, even though
    the shape is not preserved
    """
    dtype = [('x', 'f4', (nvec,))]
    num = 10
    data = np.zeros(num, dtype=dtype)
    data['x'] = np.arange(num * nvec).reshape(num, nvec)
    assert data['x'].shape == (num, nvec)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data)

            d = fits[1].read()
            if nvec == 1:
                assert d['x'].shape == (num,)
            compare_array(
                data['x'].ravel(),
                d['x'].ravel(),
                "table single field read 'x'",
            )

        # see if our convenience functions are working
        write(
            fname,
            data,
            extname="newext",
        )
        d = read(fname, ext='newext')
        if nvec == 1:
            assert d['x'].shape == (num,)
        compare_array(data['x'].ravel(), d['x'].ravel(), "table data2")

        # now test read_column
        with FITS(fname) as fits:
            d = fits[1].read_column('x')
            if nvec == 1:
                assert d.shape == (num,)
            compare_array(
                data['x'].ravel(), d.ravel(), "table single field read 'x'"
            )


@pytest.mark.parametrize('nvec', [2, 1])
def test_table_read_write_uvec1(nvec):
    """
    ensure the data for U string vec length 1 gets round-tripped, even though
    the shape is not preserved.  Also test 2 for consistency
    """

    dtype = [('string', 'U10', (nvec,))]
    num = 10
    data = np.zeros(num, dtype=dtype)
    sravel = data['string'].ravel()
    sravel[:] = ['%-10s' % i for i in range(num * nvec)]
    assert data['string'].shape == (num, nvec)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data)

            d = fits[1].read()

            if nvec == 1:
                assert d['string'].shape == (num,)

            compare_array(
                data['string'].ravel(),
                d['string'].ravel(),
                "table single field read 'string'",
            )

        # see if our convenience functions are working
        write(
            fname,
            data,
            extname="newext",
        )
        d = read(fname, ext='newext')

        if nvec == 1:
            assert d['string'].shape == (num,)
        compare_array(
            data['string'].ravel(),
            d['string'].ravel(),
            "table data2",
        )

        # now test read_column
        with FITS(fname) as fits:
            d = fits[1].read_column('string')

            if nvec == 1:
                assert d.shape == (num,)
            compare_array(
                data['string'].ravel(),
                d.ravel(),
                "table single field read 'string'",
            )


def test_table_column_index_scalar():
    """
    Test a basic table write, data and a header, then reading back in to
    check the values
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.empty(1, dtype=[('Z', 'f8')])
            data['Z'][:] = 1.0
            fits.write_table(data)
            fits.write_table(data)

        with FITS(fname, 'r') as fits:
            assert fits[1]['Z'][0].ndim == 0
            assert fits[1][0].ndim == 0


def test_table_read_empty_rows():
    """
    test reading empty list of rows from an table.
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.empty(1, dtype=[('Z', 'f8')])
            data['Z'][:] = 1.0
            fits.write_table(data)
            fits.write_table(data)

        with FITS(fname, 'r') as fits:
            assert len(fits[1].read(rows=[])) == 0
            assert len(fits[1].read(rows=range(0, 0))) == 0
            assert len(fits[1].read(rows=np.arange(0, 0))) == 0


def test_table_format_column_subset():
    """
    Test a basic table write, data and a header, then reading back in to
    check the values
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            data = np.empty(1, dtype=[('Z', 'f8'), ('Z_PERSON', 'f8')])
            data['Z'][:] = 1.0
            data['Z_PERSON'][:] = 1.0
            fits.write_table(data)
            fits.write_table(data)
            fits.write_table(data)

        with FITS(fname, 'r') as fits:
            # assert we do not have an extra row of 'Z'
            sz = str(fits[2]['Z_PERSON']).split('\n')
            s = str(fits[2][('Z_PERSON', 'Z')]).split('\n')
            assert len(sz) == len(s) - 1


def test_table_write_dict_of_arrays_scratch():
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            d = {}
            for n in data.dtype.names:
                d[n] = data[n]

            fits.write(d)

        d = read(fname)
        compare_rec(data, d, "list of dicts, scratch")


def test_table_write_dict_of_arrays():
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.create_table_hdu(data, extname='mytable')

            d = {}
            for n in data.dtype.names:
                d[n] = data[n]

            fits[-1].write(d)

        d = read(fname)
        compare_rec(data, d, "list of dicts")


def test_table_write_dict_of_arrays_var():
    """
    This version creating the table from a dict of arrays, variable
    lenght columns
    """

    adata = make_data()
    vardata = adata['vardata']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            d = {}
            for n in vardata.dtype.names:
                d[n] = vardata[n]

            fits.write(d)

        d = read(fname)
        compare_rec_with_var(vardata, d, "dict of arrays, var")


def test_table_write_list_of_arrays_scratch():
    """
    This version creating the table from the names and list, creating
    table first
    """

    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            names = [n for n in data.dtype.names]
            dlist = [data[n] for n in data.dtype.names]
            fits.write(dlist, names=names)

        d = read(fname)
        compare_rec(data, d, "list of arrays, scratch")


def test_table_write_list_of_arrays():
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.create_table_hdu(data, extname='mytable')

            columns = [n for n in data.dtype.names]
            dlist = [data[n] for n in data.dtype.names]
            fits[-1].write(dlist, columns=columns)

        d = read(fname, ext='mytable')
        compare_rec(data, d, "list of arrays")


def test_table_write_list_of_arrays_var():
    """
    This version creating the table from the names and list, variable
    lenght cols
    """
    adata = make_data()
    vardata = adata['vardata']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            names = [n for n in vardata.dtype.names]
            dlist = [vardata[n] for n in vardata.dtype.names]
            fits.write(dlist, names=names)

        d = read(fname)
        compare_rec_with_var(vardata, d, "list of arrays, var")


def test_table_write_bad_string():
    for d in ['S0', 'U0']:
        dt = [('s', d)]

        # old numpy didn't allow this dtype, so will throw
        # a TypeError for empty dtype
        try:
            data = np.zeros(1, dtype=dt)
            supported = True
        except TypeError:
            supported = False

        if supported:
            with pytest.raises(ValueError):
                with tempfile.TemporaryDirectory() as tmpdir:
                    fname = os.path.join(tmpdir, 'test.fits')
                    with FITS(fname, 'rw') as fits:
                        fits.write(data)


def test_variable_length_columns():
    adata = make_data()
    vardata = adata['vardata']

    for vstorage in ['fixed', 'object']:
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, 'test.fits')

            with FITS(fname, 'rw', vstorage=vstorage) as fits:
                fits.write(vardata)

                # reading multiple columns
                d = fits[1].read()
                compare_rec_with_var(
                    vardata, d, "read all test '%s'" % vstorage
                )

                cols = ['u2scalar', 'Sobj']
                d = fits[1].read(columns=cols)
                compare_rec_with_var(
                    vardata, d, "read all test subcols '%s'" % vstorage
                )

                # one at a time
                for f in vardata.dtype.names:
                    d = fits[1].read_column(f)
                    if util.is_object(vardata[f]):
                        compare_object_array(
                            vardata[f], d, "read all field '%s'" % f
                        )

                # same as above with slices
                # reading multiple columns
                d = fits[1][:]
                compare_rec_with_var(
                    vardata, d, "read all test '%s'" % vstorage
                )

                d = fits[1][cols][:]
                compare_rec_with_var(
                    vardata, d, "read all test subcols '%s'" % vstorage
                )

                # one at a time
                for f in vardata.dtype.names:
                    d = fits[1][f][:]
                    if util.is_object(vardata[f]):
                        compare_object_array(
                            vardata[f], d, "read all field '%s'" % f
                        )

                #
                # now same with sub rows
                #

                # reading multiple columns, sorted and unsorted
                for rows in [[0, 2], [2, 0]]:
                    d = fits[1].read(rows=rows)
                    compare_rec_with_var(
                        vardata,
                        d,
                        "read subrows test '%s'" % vstorage,
                        rows=rows,
                    )

                    d = fits[1].read(columns=cols, rows=rows)
                    compare_rec_with_var(
                        vardata,
                        d,
                        "read subrows test subcols '%s'" % vstorage,
                        rows=rows,
                    )

                    # one at a time
                    for f in vardata.dtype.names:
                        d = fits[1].read_column(f, rows=rows)
                        if util.is_object(vardata[f]):
                            compare_object_array(
                                vardata[f],
                                d,
                                "read subrows field '%s'" % f,
                                rows=rows,
                            )

                    # same as above with slices
                    # reading multiple columns
                    d = fits[1][rows]
                    compare_rec_with_var(
                        vardata,
                        d,
                        "read subrows slice test '%s'" % vstorage,
                        rows=rows,
                    )
                    d = fits[1][2:4]
                    compare_rec_with_var(
                        vardata,
                        d,
                        "read slice test '%s'" % vstorage,
                        rows=[2, 3],
                    )

                    d = fits[1][cols][rows]
                    compare_rec_with_var(
                        vardata,
                        d,
                        "read subcols subrows slice test '%s'" % vstorage,
                        rows=rows,
                    )

                    d = fits[1][cols][2:4]

                    compare_rec_with_var(
                        vardata,
                        d,
                        "read subcols slice test '%s'" % vstorage,
                        rows=[2, 3],
                    )

                    # one at a time
                    for f in vardata.dtype.names:
                        d = fits[1][f][rows]
                        if util.is_object(vardata[f]):
                            compare_object_array(
                                vardata[f],
                                d,
                                "read subrows field '%s'" % f,
                                rows=rows,
                            )
                        d = fits[1][f][2:4]
                        if util.is_object(vardata[f]):
                            compare_object_array(
                                vardata[f],
                                d,
                                "read slice field '%s'" % f,
                                rows=[2, 3],
                            )


def test_table_iter():
    """
    Test iterating over rows of a table
    """

    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data, header=adata['keys'], extname='mytable')

        # one row at a time
        with FITS(fname) as fits:
            hdu = fits["mytable"]
            i = 0
            for row_data in hdu:
                compare_rec(data[i], row_data, "table data")
                i += 1


def test_ascii_table_write_read():
    """
    Test write and read for an ascii table
    """

    adata = make_data()
    ascii_data = adata['ascii_data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(
                ascii_data,
                table_type='ascii',
                header=adata['keys'],
                extname='mytable',
            )

            # cfitsio always reports type as i4 and f8, period, even if if
            # written with higher precision.  Need to fix that somehow
            for f in ascii_data.dtype.names:
                d = fits[1].read_column(f)
                if d.dtype == np.float64:
                    # note we should be able to do 1.11e-16 in principle, but
                    # in practice we get more like 2.15e-16
                    compare_array_tol(
                        ascii_data[f], d, 2.15e-16, "table field read '%s'" % f
                    )
                else:
                    compare_array(
                        ascii_data[f], d, "table field read '%s'" % f
                    )

            for rows in [[1, 3], [3, 1]]:
                for f in ascii_data.dtype.names:
                    d = fits[1].read_column(f, rows=rows)
                    if d.dtype == np.float64:
                        compare_array_tol(
                            ascii_data[f][rows],
                            d,
                            2.15e-16,
                            "table field read subrows '%s'" % f,
                        )
                    else:
                        compare_array(
                            ascii_data[f][rows],
                            d,
                            "table field read subrows '%s'" % f,
                        )

            beg = 1
            end = 3
            for f in ascii_data.dtype.names:
                d = fits[1][f][beg:end]
                if d.dtype == np.float64:
                    compare_array_tol(
                        ascii_data[f][beg:end],
                        d,
                        2.15e-16,
                        "table field read slice '%s'" % f,
                    )
                else:
                    compare_array(
                        ascii_data[f][beg:end],
                        d,
                        "table field read slice '%s'" % f,
                    )

            cols = ['i2scalar', 'f4scalar']
            for f in ascii_data.dtype.names:
                data = fits[1].read(columns=cols)
                for f in data.dtype.names:
                    d = data[f]
                    if d.dtype == np.float64:
                        compare_array_tol(
                            ascii_data[f],
                            d,
                            2.15e-16,
                            "table subcol, '%s'" % f,
                        )
                    else:
                        compare_array(
                            ascii_data[f], d, "table subcol, '%s'" % f
                        )

                data = fits[1][cols][:]
                for f in data.dtype.names:
                    d = data[f]
                    if d.dtype == np.float64:
                        compare_array_tol(
                            ascii_data[f],
                            d,
                            2.15e-16,
                            "table subcol, '%s'" % f,
                        )
                    else:
                        compare_array(
                            ascii_data[f], d, "table subcol, '%s'" % f
                        )

            for rows in [[1, 3], [3, 1]]:
                for f in ascii_data.dtype.names:
                    data = fits[1].read(columns=cols, rows=rows)
                    for f in data.dtype.names:
                        d = data[f]
                        if d.dtype == np.float64:
                            compare_array_tol(
                                ascii_data[f][rows],
                                d,
                                2.15e-16,
                                "table subcol, '%s'" % f,
                            )
                        else:
                            compare_array(
                                ascii_data[f][rows],
                                d,
                                "table subcol, '%s'" % f,
                            )

                    data = fits[1][cols][rows]
                    for f in data.dtype.names:
                        d = data[f]
                        if d.dtype == np.float64:
                            compare_array_tol(
                                ascii_data[f][rows],
                                d,
                                2.15e-16,
                                "table subcol/row, '%s'" % f,
                            )
                        else:
                            compare_array(
                                ascii_data[f][rows],
                                d,
                                "table subcol/row, '%s'" % f,
                            )

            for f in ascii_data.dtype.names:
                data = fits[1][cols][beg:end]
                for f in data.dtype.names:
                    d = data[f]
                    if d.dtype == np.float64:
                        compare_array_tol(
                            ascii_data[f][beg:end],
                            d,
                            2.15e-16,
                            "table subcol/slice, '%s'" % f,
                        )
                    else:
                        compare_array(
                            ascii_data[f][beg:end],
                            d,
                            "table subcol/slice, '%s'" % f,
                        )


def test_table_insert_column():
    """
    Insert a new column
    """
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data, header=adata['keys'], extname='mytable')

            d = fits[1].read()

            for n in d.dtype.names:
                newname = n + '_insert'

                fits[1].insert_column(newname, d[n])

                newdata = fits[1][newname][:]

                compare_array(
                    d[n],
                    newdata,
                    "table single field insert and read '%s'" % n,
                )


def test_table_delete_row_range():
    """
    Test deleting a range of rows using the delete_rows method
    """

    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data)

        rowslice = slice(1, 3)
        with FITS(fname, 'rw') as fits:
            fits[1].delete_rows(rowslice)

        with FITS(fname) as fits:
            d = fits[1].read()

        compare_data = data[[0, 3]]
        compare_rec(compare_data, d, "delete row range")


def test_table_delete_rows():
    """
    Test deleting specific set of rows using the delete_rows method
    """

    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data)

        rows2delete = [1, 3]
        with FITS(fname, 'rw') as fits:
            fits[1].delete_rows(rows2delete)

        with FITS(fname) as fits:
            d = fits[1].read()

        compare_data = data[[0, 2]]
        compare_rec(compare_data, d, "delete rows")


def test_table_where():
    """
    Use the where method to get indices for a row filter expression
    """

    adata = make_data()
    data2 = adata['data2']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data2)

        #
        # get all indices
        #
        with FITS(fname) as fits:
            a = fits[1].where('x > 3 && y < 8')
        b = np.where((data2['x'] > 3) & (data2['y'] < 8))[0]
        np.testing.assert_array_equal(a, b)

        #
        # get slice of indices
        #
        with FITS(fname) as fits:
            a = fits[1].where('x > 3 && y < 8', 2, 8)
        b = np.where((data2['x'][2:8] > 3) & (data2['y'][2:8] < 8))[0]
        np.testing.assert_array_equal(a, b)


def test_table_resize():
    """
    Use the resize method to change the size of a table

    default values get filled in and these are tested
    """
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        #
        # shrink from back
        #
        with FITS(fname, 'rw', clobber=True) as fits:
            fits.write_table(data)

        nrows = 2
        with FITS(fname, 'rw') as fits:
            fits[1].resize(nrows)

        with FITS(fname) as fits:
            d = fits[1].read()

        compare_data = data[0:nrows]
        compare_rec(compare_data, d, "shrink from back")

        #
        # shrink from front
        #
        with FITS(fname, 'rw', clobber=True) as fits:
            fits.write_table(data)

        with FITS(fname, 'rw') as fits:
            fits[1].resize(nrows, front=True)

        with FITS(fname) as fits:
            d = fits[1].read()

        compare_data = data[nrows - data.size :]
        compare_rec(compare_data, d, "shrink from front")

        # These don't get zerod
        # the defaults below come out of cfitsio
        # IDK where they are defined
        nrows = 10
        add_data = np.zeros(nrows - data.size, dtype=data.dtype)
        add_data['i1scalar'] = -(2**7)
        add_data['i1vec'] = -(2**7)
        add_data['i1arr'] = -(2**7)
        add_data['u2scalar'] = 2**15
        add_data['u2vec'] = 2**15
        add_data['u2arr'] = 2**15
        add_data['u4scalar'] = 2**31
        add_data['u4vec'] = 2**31
        add_data['u4arr'] = 2**31
        if CFITSIO_VERSION > 4:
            add_data['u8scalar'] = 2**63
            add_data['u8vec'] = 2**63
            add_data['u8arr'] = 2**63

        #
        # expand at the back
        #
        with FITS(fname, 'rw', clobber=True) as fits:
            fits.write_table(data)
        with FITS(fname, 'rw') as fits:
            fits[1].resize(nrows)

        with FITS(fname) as fits:
            d = fits[1].read()

        compare_data = np.hstack((data, add_data))
        compare_rec(compare_data, d, "expand at the back")

        #
        # expand at the front
        #
        with FITS(fname, 'rw', clobber=True) as fits:
            fits.write_table(data)
        with FITS(fname, 'rw') as fits:
            fits[1].resize(nrows, front=True)

        with FITS(fname) as fits:
            d = fits[1].read()

        compare_data = np.hstack((add_data, data))
        # These don't get zerod
        compare_rec(compare_data, d, "expand at the front")


def test_slice():
    """
    Test reading by slice
    """
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            # initial write
            fits.write_table(data)

            # test reading single columns
            for f in data.dtype.names:
                d = fits[1][f][:]
                compare_array(
                    data[f], d, "test read all rows %s column subset" % f
                )

            # test reading row subsets
            rows = [1, 3]
            for f in data.dtype.names:
                d = fits[1][f][rows]
                compare_array(data[f][rows], d, "test %s row subset" % f)
            for f in data.dtype.names:
                d = fits[1][f][1:3]
                compare_array(data[f][1:3], d, "test %s row slice" % f)
            for f in data.dtype.names:
                d = fits[1][f][1:4:2]
                compare_array(
                    data[f][1:4:2], d, "test %s row slice with step" % f
                )
            for f in data.dtype.names:
                d = fits[1][f][::2]
                compare_array(
                    data[f][::2], d, "test %s row slice with only setp" % f
                )

            # now list of columns
            cols = ['u2scalar', 'f4vec', 'Sarr']
            d = fits[1][cols][:]
            for f in d.dtype.names:
                compare_array(data[f][:], d[f], "test column list %s" % f)

            cols = ['u2scalar', 'f4vec', 'Sarr']
            d = fits[1][cols][rows]
            for f in d.dtype.names:
                compare_array(
                    data[f][rows], d[f], "test column list %s row subset" % f
                )

            cols = ['u2scalar', 'f4vec', 'Sarr']
            d = fits[1][cols][1:3]
            for f in d.dtype.names:
                compare_array(
                    data[f][1:3], d[f], "test column list %s row slice" % f
                )


def test_table_append():
    """
    Test creating a table and appending new rows.
    """
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            # initial write
            fits.write_table(data, header=adata['keys'], extname='mytable')
            # now append
            data2 = data.copy()
            data2['f4scalar'] = 3
            fits[1].append(data2)

            d = fits[1].read()
            assert d.size == data.size * 2

            compare_rec(data, d[0 : data.size], "Comparing initial write")
            compare_rec(data2, d[data.size :], "Comparing appended data")

            h = fits[1].read_header()
            compare_headerlist_header(adata['keys'], h)

            # append with list of arrays and names
            names = data.dtype.names
            data3 = [np.array(data[name]) for name in names]
            fits[1].append(data3, names=names)

            d = fits[1].read()
            assert d.size == data.size * 3
            compare_rec(data, d[2 * data.size :], "Comparing appended data")

            # append with list of arrays and columns
            fits[1].append(data3, columns=names)

            d = fits[1].read()
            assert d.size == data.size * 4
            compare_rec(data, d[3 * data.size :], "Comparing appended data")


def test_table_subsets():
    """
    testing reading subsets
    """
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data, header=adata['keys'], extname='mytable')

            for rows in [[1, 3], [3, 1]]:
                d = fits[1].read(rows=rows)
                compare_rec_subrows(data, d, rows, "table subset")
                columns = ['i1scalar', 'f4arr']
                d = fits[1].read(columns=columns, rows=rows)

                for f in columns:
                    d = fits[1].read_column(f, rows=rows)
                    compare_array(
                        data[f][rows], d, "row subset, multi-column '%s'" % f
                    )
                for f in data.dtype.names:
                    d = fits[1].read_column(f, rows=rows)
                    compare_array(
                        data[f][rows], d, "row subset, column '%s'" % f
                    )


def test_gz_write_read():
    """
    Test a basic table write, data and a header, then reading back in to
    check the values

    this code all works, but the file is zere size when done!
    """
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data, header=adata['keys'], extname='mytable')

            d = fits[1].read()
            compare_rec(data, d, "gzip write/read")

            h = fits[1].read_header()
            for entry in adata['keys']:
                name = entry['name'].upper()
                value = entry['value']
                hvalue = h[name]
                if isinstance(hvalue, str):
                    hvalue = hvalue.strip()
                assert value == hvalue, "testing header key '%s'" % name

                if 'comment' in entry:
                    assert (
                        entry['comment'].strip() == h.get_comment(name).strip()
                    ), "testing comment for header key '%s'" % name

        stat = os.stat(fname)
        assert stat.st_size != 0, "Making sure the data was flushed to disk"


@pytest.mark.skipif(
    not cfitsio_has_bzip2_support(),
    reason='cfitsio was not built with bzip2 support',
)
def test_bz2_read():
    '''
    Write a normal .fits file, run bzip2 on it, then read the bz2
    file and verify that it's the same as what we put in; we don't
    [currently support or] test *writing* bzip2.
    '''

    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        bzfname = fname + '.bz2'

        try:
            fits = FITS(fname, 'rw')
            fits.write_table(data, header=adata['keys'], extname='mytable')
            fits.close()

            os.system('bzip2 %s' % fname)
            f2 = FITS(bzfname)
            d = f2[1].read()
            compare_rec(data, d, "bzip2 read")

            h = f2[1].read_header()
            for entry in adata['keys']:
                name = entry['name'].upper()
                value = entry['value']
                hvalue = h[name]
                if isinstance(hvalue, str):
                    hvalue = hvalue.strip()

                assert value == hvalue, "testing header key '%s'" % name

                if 'comment' in entry:
                    assert (
                        entry['comment'].strip() == h.get_comment(name).strip()
                    ), "testing comment for header key '%s'" % name
        except Exception:
            import traceback

            traceback.print_exc()

            assert False, 'Exception in testing bzip2 reading'


def test_checksum():
    """
    test that checksumming works
    """
    adata = make_data()
    data = adata['data']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(data, header=adata['keys'], extname='mytable')
            fits[1].write_checksum()
            fits[1].verify_checksum()


def test_trim_strings():
    """
    test mode where we strim strings on read
    """

    dt = [('fval', 'f8'), ('name', 'S15'), ('vec', 'f4', 2)]
    n = 3
    data = np.zeros(n, dtype=dt)
    data['fval'] = np.random.random(n)
    data['vec'] = np.random.random(n * 2).reshape(n, 2)

    data['name'] = ['mike', 'really_long_name_to_fill', 'jan']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write(data)

        for onconstruct in [True, False]:
            if onconstruct:
                ctrim = True
                otrim = False
            else:
                ctrim = False
                otrim = True

            with FITS(fname, 'rw', trim_strings=ctrim) as fits:
                if ctrim:
                    dread = fits[1][:]
                    compare_rec(
                        data,
                        dread,
                        "trimmed strings constructor",
                    )

                    dname = fits[1]['name'][:]
                    compare_array(
                        data['name'],
                        dname,
                        "trimmed strings col read, constructor",
                    )
                    dread = fits[1][['name']][:]
                    compare_array(
                        data['name'],
                        dread['name'],
                        "trimmed strings col read, constructor",
                    )

                dread = fits[1].read(trim_strings=otrim)
                compare_rec(
                    data,
                    dread,
                    "trimmed strings keyword",
                )
                dname = fits[1].read(columns='name', trim_strings=otrim)
                compare_array(
                    data['name'],
                    dname,
                    "trimmed strings col keyword",
                )
                dread = fits[1].read(columns=['name'], trim_strings=otrim)
                compare_array(
                    data['name'],
                    dread['name'],
                    "trimmed strings col keyword",
                )

        # convenience function
        dread = read(fname, trim_strings=True)
        compare_rec(
            data,
            dread,
            "trimmed strings convenience function",
        )
        dname = read(fname, columns='name', trim_strings=True)
        compare_array(
            data['name'],
            dname,
            "trimmed strings col convenience function",
        )
        dread = read(fname, columns=['name'], trim_strings=True)
        compare_array(
            data['name'],
            dread['name'],
            "trimmed strings col convenience function",
        )


def test_lower_upper():
    """
    test forcing names to upper and lower
    """

    rng = np.random.RandomState(8908)

    dt = [('MyName', 'f8'), ('StuffThings', 'i4'), ('Blah', 'f4')]
    data = np.zeros(3, dtype=dt)
    data['MyName'] = rng.uniform(data.size)
    data['StuffThings'] = rng.uniform(data.size)
    data['Blah'] = rng.uniform(data.size)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write(data)

        for i in [1, 2]:
            if i == 1:
                lower = True
                upper = False
            else:
                lower = False
                upper = True

            with FITS(fname, 'rw', lower=lower, upper=upper) as fits:
                for rows in [None, [1, 2]]:
                    d = fits[1].read(rows=rows)
                    compare_names(
                        d.dtype.names,
                        data.dtype.names,
                        lower=lower,
                        upper=upper,
                    )

                    d = fits[1].read(
                        rows=rows, columns=['MyName', 'stuffthings']
                    )
                    compare_names(
                        d.dtype.names,
                        data.dtype.names[0:2],
                        lower=lower,
                        upper=upper,
                    )

                    d = fits[1][1:2]
                    compare_names(
                        d.dtype.names,
                        data.dtype.names,
                        lower=lower,
                        upper=upper,
                    )

                    if rows is not None:
                        d = fits[1][rows]
                    else:
                        d = fits[1][:]

                    compare_names(
                        d.dtype.names,
                        data.dtype.names,
                        lower=lower,
                        upper=upper,
                    )

                    if rows is not None:
                        d = fits[1][['myname', 'stuffthings']][rows]
                    else:
                        d = fits[1][['myname', 'stuffthings']][:]

                    compare_names(
                        d.dtype.names,
                        data.dtype.names[0:2],
                        lower=lower,
                        upper=upper,
                    )

            # using overrides
            with FITS(fname, 'rw') as fits:
                for rows in [None, [1, 2]]:
                    d = fits[1].read(rows=rows, lower=lower, upper=upper)
                    compare_names(
                        d.dtype.names,
                        data.dtype.names,
                        lower=lower,
                        upper=upper,
                    )

                    d = fits[1].read(
                        rows=rows,
                        columns=['MyName', 'stuffthings'],
                        lower=lower,
                        upper=upper,
                    )
                    compare_names(
                        d.dtype.names,
                        data.dtype.names[0:2],
                        lower=lower,
                        upper=upper,
                    )

            for rows in [None, [1, 2]]:
                d = read(fname, rows=rows, lower=lower, upper=upper)
                compare_names(
                    d.dtype.names, data.dtype.names, lower=lower, upper=upper
                )

                d = read(
                    fname,
                    rows=rows,
                    columns=['MyName', 'stuffthings'],
                    lower=lower,
                    upper=upper,
                )
                compare_names(
                    d.dtype.names,
                    data.dtype.names[0:2],
                    lower=lower,
                    upper=upper,
                )


def test_read_raw():
    """
    testing reading the file as raw bytes
    """
    rng = np.random.RandomState(8908)

    dt = [('MyName', 'f8'), ('StuffThings', 'i4'), ('Blah', 'f4')]
    data = np.zeros(3, dtype=dt)
    data['MyName'] = rng.uniform(data.size)
    data['StuffThings'] = rng.uniform(data.size)
    data['Blah'] = rng.uniform(data.size)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        try:
            with FITS(fname, 'rw') as fits:
                fits.write(data)
                raw1 = fits.read_raw()

            with FITS('mem://', 'rw') as fits:
                fits.write(data)
                raw2 = fits.read_raw()

            with open(fname, 'rb') as fobj:
                raw3 = fobj.read()

            assert raw1 == raw2
            assert raw1 == raw3
        except Exception:
            import traceback

            traceback.print_exc()
            assert False, 'Exception in testing read_raw'


def test_table_bitcol_read_write():
    """
    Test basic write/read with bitcols
    """

    adata = make_data()
    bdata = adata['bdata']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.write_table(bdata, extname='mytable', write_bitcols=True)

            d = fits[1].read()
            compare_rec(bdata, d, "table read/write")

            rows = [0, 2]
            d = fits[1].read(rows=rows)
            compare_rec(bdata[rows], d, "table read/write rows")

            d = fits[1][:2]
            compare_rec(bdata[:2], d, "table read/write slice")

        # now test read_column
        with FITS(fname) as fits:
            for f in bdata.dtype.names:
                d = fits[1].read_column(f)
                compare_array(
                    bdata[f], d, "table 1 single field read '%s'" % f
                )

            # now list of columns
            for cols in [['b1vec', 'b1arr']]:
                d = fits[1].read(columns=cols)
                for f in d.dtype.names:
                    compare_array(bdata[f][:], d[f], "test column list %s" % f)

                for rows in [[1, 3], [3, 1]]:
                    d = fits[1].read(columns=cols, rows=rows)
                    for f in d.dtype.names:
                        compare_array(
                            bdata[f][rows],
                            d[f],
                            "test column list %s row subset" % f,
                        )


def test_table_bitcol_append():
    """
    Test creating a table with bitcol support and appending new rows.
    """
    adata = make_data()
    bdata = adata['bdata']

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            # initial write
            fits.write_table(bdata, extname='mytable', write_bitcols=True)

        with FITS(fname, 'rw') as fits:
            # now append
            bdata2 = bdata.copy()
            fits[1].append(bdata2)

            d = fits[1].read()
            assert d.size == bdata.size * 2

            compare_rec(bdata, d[0 : bdata.size], "Comparing initial write")
            compare_rec(bdata2, d[bdata.size :], "Comparing appended data")


def test_table_bitcol_insert():
    """
    Test creating a table with bitcol support and appending new rows.
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            # initial write
            nrows = 3
            d = np.zeros(nrows, dtype=[('ra', 'f8')])
            d['ra'] = range(d.size)
            fits.write(d)

        with FITS(fname, 'rw') as fits:
            bcol = np.array([True, False, True])

            # now append
            fits[-1].insert_column(
                'bscalar_inserted', bcol, write_bitcols=True
            )

            d = fits[-1].read()
            assert d.size == nrows, 'read size equals'
            compare_array(bcol, d['bscalar_inserted'], "inserted bitcol")

            bvec = np.array([[True, False], [False, True], [True, True]])

            # now append
            fits[-1].insert_column('bvec_inserted', bvec, write_bitcols=True)

            d = fits[-1].read()
            assert d.size == nrows, 'read size equals'
            compare_array(bvec, d['bvec_inserted'], "inserted bitcol")


def test_table_write_dict_of_arrays_unaligned():
    data = {}
    for dtype in DTYPES:
        _data = np.arange(20, dtype=dtype)
        # The code to make the unaligned view was generated
        # by Google's AI and then modified by hand to fix a bug.
        unaligned_data = np.ndarray(
            shape=(19,),
            dtype=_data.dtype,
            buffer=_data.data,
            offset=1,  # Offset by 1 byte
            strides=_data.strides,
        )
        if not dtype.endswith("1"):
            assert not unaligned_data.flags["ALIGNED"]

        data[dtype.replace("<", "l")] = unaligned_data

    dtype = np.dtype(
        {
            "names": list(data.keys()),
            "formats": [v.dtype for v in data.values()],
        }
    )
    data_stra = np.zeros(data[dtype.names[0]].shape, dtype=dtype)
    for k, v in data.items():
        data_stra[k] = v

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            fits.create_table_hdu(data, extname='mytable')
            fits[-1].write(data)

        d = read(fname)
        compare_rec(data_stra, d, "list of dicts")


@pytest.mark.parametrize("table_type", ["binary", "ascii"])
def test_table_big_col(table_type):
    d = np.ones(1, dtype=[("blah", "U70000")])
    d["blah"] = "".join(["a"] * 60000)
    with tempfile.TemporaryDirectory() as tmpdir:
        pth = os.path.join(tmpdir, "test.fits")
        # v3 cfitsio that is not bundled fails for big
        # columns
        if table_type == "ascii" or CFITSIO_VERSION < 4:
            with pytest.raises(OSError) as e:
                write(pth, d, table_type=table_type)
            assert "FITSIO status = 236: column exceeds width of table" in str(
                e.value
            )
            assert (
                "string column is too wide: 70000; "
                "max supported width is" in str(e.value)
            )
        else:
            write(pth, d, table_type=table_type)
            data = read(pth)
            np.testing.assert_array_equal(d, data)


@pytest.mark.xfail(
    condition=CFITSIO_VERSION < 4,
    reason=(
        "cfitsio versions < 4 do not easily support null-terminated strings"
    ),
)
@pytest.mark.parametrize("table_type", ["binary", "ascii"])
def test_table_null_end_strings(table_type):
    d = np.ones(2, dtype=[("blah", "U70")])
    d["blah"][0] = "".join(["a"] * 60)
    d["blah"][1] = ""
    with tempfile.TemporaryDirectory() as tmpdir:
        pth = os.path.join(tmpdir, "test.fits")
        write(pth, d, table_type=table_type)
        data = read(pth)
        assert len(data["blah"][0]) == 60
        assert "U70" in data["blah"].dtype.descr[0][1]

        if table_type == "ascii":
            # null strings in ascii tables are a single blank
            d["blah"][1] = " "
        np.testing.assert_array_equal(d, data)


def test_table_read_write_ulonglong():
    adata = np.zeros(5, dtype=[("u8scalar", "u8")])
    adata["u8scalar"] = (2**64 - 1) - np.arange(5, dtype="u8")

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            if CFITSIO_VERSION < 3.45:
                with pytest.raises(IOError) as e:
                    fits.write_table(
                        adata,
                        extname='mytable',
                    )
                assert "'W'" in str(e.value)
            else:
                fits.write_table(
                    adata,
                    extname='mytable',
                )
                d = fits[1].read()
                compare_rec(adata, d, "table read/write")


@pytest.mark.parametrize("typ", ["u8", "u4", "i8"])
def test_table_read_write_ulonglong_ascii_raises(typ):
    adata = np.zeros(5, dtype=[("scalar", typ)])
    if typ == "u8":
        val = 2**64 - 1
    elif typ == "u4":
        val = 2**32 - 1
    elif typ == "i8":
        val = 2**31 - 1
    adata["scalar"] = (val) - np.arange(5, dtype=typ)

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'test.fits')

        with FITS(fname, 'rw') as fits:
            with pytest.raises(ValueError) as e:
                fits.write_table(
                    adata,
                    extname='mytable',
                    table_type='ascii',
                )
            assert f"unsupported type '{typ}' for ascii tables" in str(e.value)
